#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Union, Tuple

from flsim.interfaces.batch_metrics import IFLBatchMetrics
from torch import Tensor


class Metric:
    r"""
    Wraps a metric.

    A reportable metric is simply the name of the metric,
    and the value of the metric, in its simplest form.
    The value could also be a dict of other metrics, which
    in this case the metric is a set of other metrics, and
    the `is_compound` attribute is set.
    """

    def __init__(self, name: str, value: Union[float, List["Metric"]]):
        self.name = name
        self.value = value

    @property
    def is_compund(self):
        return isinstance(self.value, list)

    def __str__(self):
        return f"{self.name}: {self.value}"

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> List["Metric"]:

        metrics = []

        def process_dict(d, metrics):
            for k, v in d.items():
                assert isinstance(k, str), f"{k} must be a string"
                if isinstance(v, dict):
                    sub_metric = []
                    process_dict(v, sub_metric)
                    metrics.append(Metric(k, sub_metric, True))
                else:
                    assert isinstance(
                        v, (int, float, Tensor)
                    ), f"{v} is not of types int, float, or torch.Tensor"
                    metrics.append(Metric(k, float(v)))

        process_dict(d, metrics)
        return metrics

    @classmethod
    def from_args(cls, **kwargs):
        r"""
        Simple util to generate Metrics from kwargs.

        The usage is simple, metrics need to be passed as named arguments
        to the function. The class will throw if the metrics are not
        any of the valid types: int, float, tensor of size 1, or a
        dictionary of such types with string keys. This latter case is considerd
        a metric with sub metrics.

        Example:
        metric = Metrics.from_args(a=1, b=2.0, c={d=1, e=2})
        will result in:

        [Metric(a, 1.0), Metric(b, 2.0), Metric(c, [Metric(d, 1.0), Metric(e, 2.0)])]
        """
        return cls.from_dict(kwargs)

    @classmethod
    def to_dict(cls, metrics):

        d = {}

        def process_list(metrics, d):
            for metric in metrics:
                assert isinstance(metric, Metric)
                value = metric.value
                if metric.is_compund:
                    value = {}
                    process_list(metric.value, value)
                d[metric.name] = value

        process_list(metrics, d)
        return d


class Channel(Enum):
    """Enum that indicates to which "channel" we'd want to log/print our
    metrics to. For example, if on chooses `Channel.TENSORBOARD`, a
    metrics reporter would display the information on the TensorBoard.
    """

    TENSORBOARD = auto()
    STDOUT = auto()


class TrainingStage(Enum):
    """Enum that indicates at which stage training is, which will be used when
    reporting metrics to certain channels.
    """

    TRAINING = auto()
    AGGREGATION = auto()
    TEST = auto()
    EVAL = auto()
    PER_CLIENT_EVAL = auto()


class IFLMetricsReporter(abc.ABC):
    """Interface that all PyML FLMetricsReporter should implement. Each user
    will have 1 reporter throughout one's entire training. At the beginning of
    user’s training (i.e. at the start of epoch for a user), the user starts by
    clearing up any aggregation left by calling reset(). After the batch, each
    user collects one’s own metrics by calling add_batch_metrics() method. When
    all the batches are completed (i.e. at the end of all local epochs for a user),
    a global MetricsReporter (i.e. the MetricsReporter responsible of the whole
    training aggregates all the data by calling aggregate() method, which gets
    a MetricsReporter of an user who just completed one’s own epoch. Then, after
    all users’ local epochs are completed, the global MetricsReporter completes
    its global aggreation and report its metrics to given channels for that global
    epoch.
    Note: 1 global epoch consists of several rounds. In each round, we train
    a subset of users and each user goes through a number of local epochs, where
    each local epoch consists of multiple batches.
    """

    @abc.abstractmethod
    def add_batch_metrics(self, metrics: IFLBatchMetrics) -> None:
        """Take in output of training for a batch (of each user).
        Aggregates metrics (e.g. accuracy, loss, predictions, etc) from batch
        into state.
        """
        pass

    @abc.abstractmethod
    def aggregate(self, one_user_metrics: "IFLMetricsReporter"):
        """Combine metrics from one user into a global metrics."""
        pass

    @abc.abstractmethod
    def report_metrics(
        self,
        reset: bool,
        stage: TrainingStage,
        extra_metrics: Optional[List[Metric]] = None,
        **kwargs,
    ) -> Tuple[Any, bool]:
        """Report metrics to certain channels such as stdout, file, TensorBoard,
        etc. Also, one may want to reset metrics if needed after reporting.
        Return value: A tuple with two elements:
            1. A metrics object
            2. bool: Were the best eval metrics updated?
               Current eval metrics are compared to the best eval metrics. If
               current eval metrics are better than best eval metrics, true is returned.
               Comparing eval metrics to best eval metrics is common in
               many ML training algorithms, e.g, early stopping.
        """
        pass

    # TODO: is this needed? Do we ever call this externally?
    @abc.abstractmethod
    def reset(self):
        """Clean up all aggregations so far."""
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple

from flsim.channels.base_channel import IdentityChannel
from flsim.interfaces.model import IFLModel
from flsim.reducers.base_round_reducer import RoundReducer, RoundReducerConfig
from flsim.secure_aggregation.secure_aggregator import (
    FixedPointConfig,
    SecureAggregator,
    utility_config_flatter,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from torch import nn


class SecureRoundReducer(RoundReducer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        num_users_per_round: Optional[int] = None,
        total_number_of_users: Optional[int] = None,
        channel: Optional[IdentityChannel] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SecureRoundReducerConfig,
            **kwargs,
        )

        super().__init__(
            global_model=global_model,
            num_users_per_round=num_users_per_round,
            total_number_of_users=total_number_of_users,
            channel=channel,
            name=name,
            **kwargs,
        )
        # pyre-fixme[16]: `SecureRoundReducer` has no attribute `cfg`.
        self.sec_agg_on = self.cfg.fixedpoint is not None
        if self.sec_agg_on:
            self.secure_aggregator = self._init_secure_aggregator(
                self.cfg.fixedpoint, global_model
            )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _init_secure_aggregator(
        self,
        fixedpoint: FixedPointConfig,
        global_model: IFLModel,
    ) -> SecureAggregator:
        """
        Initializes a secure aggregation object based on the received config

        Args:
            fixedpoint: config for fixedpoint
            global_model: the global model

        Returns:
            A secure aggregation object based on the config

        Notes:
            Once we allow per-layer config for FixedPoint, we should change the
            fixedpoint to a dictionary or a similar data structure. T97188526
        """
        return SecureAggregator(
            utility_config_flatter(
                global_model.fl_get_module(),
                fixedpoint,
            )
        )

    def reduce(self) -> Tuple[nn.Module, float]:
        if not self.sec_agg_on:
            return super().reduce()
        self._reduce_all()
        # Apply secAgg operations related to model aggregate
        self.secure_aggregator.apply_denoise_mask(
            self.reduced_module.named_parameters()
        )
        self.secure_aggregator.params_to_float(self.reduced_module)
        return self.reduced_module, float(self.sum_weights.item())

    def update_reduced_module(self, delta_module: nn.Module, weight: float) -> None:
        if not self.sec_agg_on:
            return super().update_reduced_module(delta_module, weight)
        # TODO num_samples is used as the default weight, this needs revisit
        if not self.is_weighted:
            weight = 1.0
        # Apply secAgg operations related to model updates and
        # reduced_module, which is the model aggregate.
        FLModelParamUtils.multiply_model_by_weight(
            model=delta_module,
            weight=weight,
            model_to_save=delta_module,
            # pyre-fixme[16]: `SecureRoundReducer` has no attribute `cfg`.
            only_federated_params=self.cfg.only_federated_params,
        )  # first, apply weight to delta, to get (delta * weight)

        self.secure_aggregator.params_to_fixedpoint(delta_module)
        self.secure_aggregator.apply_noise_mask(delta_module.named_parameters())
        FLModelParamUtils.add_model(
            model1=self.reduced_module,  # in FixedPoint
            model2=delta_module,  # in FixedPoint
            model_to_save=self.reduced_module,
            only_federated_params=self.cfg.only_federated_params,
        )
        self.sum_weights += weight


@dataclass
class SecureRoundReducerConfig(RoundReducerConfig):
    """
    Contains configurations for a round reducer with Secure Aggregation
    """

    _target_: str = fullclassname(SecureRoundReducer)
    fixedpoint: Optional[FixedPointConfig] = None



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from time import time
from typing import Any, Dict, Iterable, List, Optional, Tuple

import torch
from flsim.channels.message import Message
from flsim.clients.base_client import Client
from flsim.clients.dp_client import DPClientConfig, DPClient
from flsim.common.timeline import Timeline
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter, Metric, TrainingStage
from flsim.interfaces.model import IFLModel
from flsim.servers.sync_dp_servers import SyncDPSGDServerConfig
from flsim.servers.sync_servers import (
    ISyncServer,
    SyncServerConfig,
    FedAvgOptimizerConfig,
)
from flsim.trainers.trainer_base import FLTrainer, FLTrainerConfig
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.config_utils import is_target
from flsim.utils.distributed.fl_distributed import FLDistributedUtils
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.fl.stats import RandomVariableStatsTracker
from hydra.utils import instantiate
from omegaconf import OmegaConf
from tqdm import tqdm


class SyncTrainer(FLTrainer):
    """Implements FederatedAveraging: https://arxiv.org/abs/1602.05629

    Attributes:
        epochs (int): Training epochs
        report_train_metrics (bool): Whether metrics on training data should be
            computed and reported.
    """

    def __init__(
        self,
        *,
        model: IFLModel,
        cuda_enabled: bool = False,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SyncTrainerConfig,
            **kwargs,
        )

        super().__init__(model=model, cuda_enabled=cuda_enabled, **kwargs)
        self.server: ISyncServer = instantiate(
            # pyre-ignore[16]
            self.cfg.server,
            global_model=model,
            channel=self.channel,
        )
        self.clients = {}

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.server, "_target_"):
            cfg.server = SyncServerConfig(optimizer=FedAvgOptimizerConfig())

    def global_model(self) -> IFLModel:
        """This function makes it explicit that self.global_model() is owned
        by the aggregator, not by SyncTrainer
        """
        return self.server.global_model

    @property
    def is_user_level_dp(self):
        return is_target(self.cfg.server, SyncDPSGDServerConfig)

    @property
    def is_sample_level_dp(self):
        return is_target(self.cfg.client, DPClientConfig)

    def create_or_get_client_for_data(self, dataset_id: int, datasets: Any):
        """This function is used to create clients in a round. Thus, it
        is called UPR * num_rounds times per training run. Here, we use
        <code>OmegaConf.structured</code> instead of <code>hydra.instantiate</code>
        to minimize the overhead of hydra object creation.
        """
        if self.is_sample_level_dp:
            client = DPClient(
                # pyre-ignore[16]
                **OmegaConf.structured(self.cfg.client),
                dataset=datasets[dataset_id],
                name=f"client_{dataset_id}",
                timeout_simulator=self._timeout_simulator,
                store_last_updated_model=self.cfg.report_client_metrics,
                channel=self.channel,
                cuda_manager=self._cuda_state_manager,
            )
        else:
            client = Client(
                **OmegaConf.structured(self.cfg.client),
                dataset=datasets[dataset_id],
                name=f"client_{dataset_id}",
                timeout_simulator=self._timeout_simulator,
                store_last_updated_model=self.cfg.report_client_metrics,
                channel=self.channel,
                cuda_manager=self._cuda_state_manager,
            )
        self.clients[dataset_id] = client
        return self.clients[dataset_id]

    def train(
        self,
        data_provider: IFLDataProvider,
        metric_reporter: IFLMetricsReporter,
        num_total_users: int,
        distributed_world_size: int,
        rank: int = 0,
    ) -> Tuple[IFLModel, Any]:
        """Train and eval a model, the model states will be modified. This function
        iterates over epochs specified in config, and for each epoch:

            1. Trains model in a federated way: different models are trained over data
                from different users, and are averaged into 'model' at the end of epoch
            2. Evaluate averaged model using evaluation data
            3. Calculate metrics based on evaluation results and select best model

        Args:
            train_iter (List[Iterable[Any]]): list of batch iterators of training data,
                each batch iterator represents data from a single 'user'
            eval_iter (Iterable[Any]): batch iterator of evaluation data
            model (Model): model to be trained
            metric_reporter (IFLMetricsReporter): compute metric based on training
                output and report results to console, file.. etc
            train_config (PyTextConfig): training config

        Returns:
            model, best_metric: the trained model together with the best metric

        Note:
            one `epoch` = go over all users once is not True here
            since users in each round are selected randomly, this isn't precisely true
            we may go over some users more than once, and some users never
            however, as long as users_per_round << num_total_users, this will work
            the alternative is to keep track of all users that have already
            been selected in the current epoch - impractical and not worth it
            however, we may have that option in simulation later on.
            TODO correct note if above option added.
        """
        # set up synchronization utilities for distributed training
        FLDistributedUtils.setup_distributed_training(
            distributed_world_size, use_cuda=self.cuda_enabled
        )  # TODO do not call dsitributed utils here, this is upstream responsibility
        self.logger.info(f" dist world size = {distributed_world_size}")

        if rank != 0:
            FLDistributedUtils.suppress_output()

        # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
        assert self.cfg.users_per_round % distributed_world_size == 0

        best_metric = None
        best_model_state = self.global_model().fl_get_module().state_dict()
        # last_best_epoch = 0
        users_per_round = min(self.cfg.users_per_round, num_total_users)

        self.data_provider = data_provider
        num_rounds_in_epoch = self.rounds_in_one_epoch(num_total_users, users_per_round)
        num_users_on_worker = data_provider.num_users()
        self.logger.debug(
            f"num_users_on_worker: {num_users_on_worker}, "
            f"users_per_round: {users_per_round}, "
            f"num_total_users: {num_total_users}"
        )
        # torch.multinomial requires int instead of float, cast it as int
        users_per_round_on_worker = int(users_per_round / distributed_world_size)
        self._validate_users_per_round(users_per_round_on_worker, num_users_on_worker)

        self.logger.info("Start training")
        if self.logger.isEnabledFor(logging.DEBUG):
            norm = FLModelParamUtils.debug_model_norm(
                self.global_model().fl_get_module()
            )
            self.logger.debug(
                self.cuda_enabled and distributed_world_size > 1,
                f"from worker {rank}: model norm is {norm} after round {iter}",
            )

        # main training loop
        num_int_epochs = math.ceil(self.cfg.epochs)
        for epoch in tqdm(
            range(1, num_int_epochs + 1), desc="Epoch", unit="epoch", position=0
        ):
            for round in tqdm(
                range(1, num_rounds_in_epoch + 1),
                desc="Round",
                unit="round",
                position=0,
            ):
                timeline = Timeline(
                    epoch=epoch, round=round, rounds_per_epoch=num_rounds_in_epoch
                )

                t = time()
                clients = self._client_selection(
                    num_users_on_worker,
                    users_per_round_on_worker,
                    data_provider,
                    self.global_model(),
                    epoch,
                )
                self.logger.info(f"Client Selection took: {time() - t} s.")

                agg_metric_clients = self._choose_clients_for_post_aggregation_metrics(
                    train_clients=clients,
                    num_total_users=num_users_on_worker,
                    users_per_round=users_per_round_on_worker,
                )

                # training on selected clients for the round
                self.logger.info(f"# clients/round on worker {rank}: {len(clients)}.")
                self._train_one_round(
                    timeline=timeline,
                    clients=clients,
                    agg_metric_clients=agg_metric_clients,
                    metric_reporter=metric_reporter
                    if self.cfg.report_train_metrics
                    else None,
                )

                if self.logger.isEnabledFor(logging.DEBUG):
                    norm = FLModelParamUtils.debug_model_norm(
                        self.global_model().fl_get_module()
                    )
                    self.logger.debug(
                        self.cuda_enabled and distributed_world_size > 1,
                        f"from worker {rank}: model norm: {norm} @ "
                        f"epoch:{epoch}, round:{round}",
                    )

                # report training success rate and training time variance
                # TODO write to metrics reporter T62638873
                if rank == 0:
                    if (
                        self._timeout_simulator.sample_mean_per_user != 0
                        or self._timeout_simulator.sample_var_per_user != 0
                    ):
                        self.logger.info(
                            f"mean training time/user: "
                            f"{self._timeout_simulator.sample_mean_per_user}",
                            f"variance of training time/user: "
                            f"{self._timeout_simulator.sample_var_per_user}",
                        )

                    t = time()
                    (best_metric, best_model_state,) = self._maybe_run_evaluation(
                        self.global_model(),
                        timeline,
                        data_provider.eval_data(),
                        metric_reporter,
                        best_metric,
                        best_model_state,
                    )
                    self.logger.info(f"Evaluation took {time() - t} s.")

                if self.stop_fl_training(
                    epoch=epoch, round=round, num_rounds_in_epoch=num_rounds_in_epoch
                ):
                    break

            # pyre-fixme[61]: `timeline` may not be initialized here.
            self._post_epoch_client_metrics_eval(timeline, metric_reporter)
            if self.stop_fl_training(
                epoch=epoch,
                round=round,  # pyre-fixme[61]: `round` may not be initialized here.
                num_rounds_in_epoch=num_rounds_in_epoch,
            ):
                break

        if rank == 0 and best_metric is not None:
            self._save_model_and_metrics(self.global_model(), best_model_state)

        return self.global_model(), best_metric

    def _post_epoch_client_metrics_eval(
        self,
        timeline: Timeline,
        metric_reporter: IFLMetricsReporter,
    ):
        self._report_post_epoch_client_metrics(
            timeline=timeline,
            metric_reporter=metric_reporter,
        )

    def stop_fl_training(self, *, epoch, round, num_rounds_in_epoch) -> bool:
        # stop if necessary number of steps/epochs are completed in case of fractional epochs
        # or if client times out
        global_round_num = (epoch - 1) * num_rounds_in_epoch + round
        return (
            (global_round_num / num_rounds_in_epoch)
            >= self.cfg.epochs  # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
            or self._timeout_simulator.stop_fl()
        )

    def _drop_overselected_users(
        self, clents_triggered: List[Client], num_users_keep: int
    ) -> List[Client]:
        """
        sort users by their training time, and only keep num_users_keep users
        """
        all_training_times = [c.get_total_training_time() for c in clents_triggered]
        all_training_times.sort()
        # only select first num_users_keep userids sort by their finish time
        num_users_keep = min([num_users_keep, len(all_training_times)])
        last_user_time = all_training_times[num_users_keep - 1]
        num_users_added = 0
        clients_used = []
        for c in clents_triggered:
            # if two clients finished at the same time, order for entering
            # the cohort is arbitrary
            if (c.get_total_training_time() <= last_user_time) and (
                num_users_added < num_users_keep
            ):
                num_users_added += 1
                clients_used.append(c)

        return clients_used

    def _client_selection(
        self,
        num_users: int,
        users_per_round: int,
        data_provider: IFLDataProvider,
        global_model: IFLModel,
        epoch: int,
    ) -> List[Client]:
        # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
        num_users_overselected = math.ceil(users_per_round / self.cfg.dropout_rate)
        user_indices_overselected = self.server.select_clients_for_training(
            num_total_users=num_users,
            users_per_round=num_users_overselected,
            data_provider=data_provider,
            epoch=epoch,
        )
        clients_triggered = [
            self.create_or_get_client_for_data(i, self.data_provider)
            for i in user_indices_overselected
        ]
        clients_to_train = self._drop_overselected_users(
            clients_triggered, users_per_round
        )
        return clients_to_train

    def _save_model_and_metrics(self, model: IFLModel, best_model_state):
        model.fl_get_module().load_state_dict(best_model_state)

    def _train_one_round(
        self,
        timeline: Timeline,
        clients: Iterable[Client],
        agg_metric_clients: Iterable[Client],
        metric_reporter: Optional[IFLMetricsReporter],
    ) -> None:
        """Args:
        timeline: information about the round, epoch, round number, ...
        clients: clients for the round
        agg_metric_clients: clients for evaluating the post-aggregation training metrics
        metric_reporter: the metric reporter to pass to other methods
        """
        t = time()
        self.server.init_round()
        self.logger.info(f"Round initialization took {time() - t} s.")

        def update(client):
            client_delta, weight = client.generate_local_update(
                self.global_model(), metric_reporter
            )
            self.server.receive_update_from_client(Message(client_delta, weight))

        # TODO MM add multi-processing here for now we will do sequential
        # with Pool(100) as workers:
        #    workers.map(client_update, clients)

        t = time()
        [update(client) for client in clients]
        self.logger.info(f"Collecting round's clients took {time() - t} s.")

        t = time()
        self.server.step()
        self.logger.info(f"Finalizing round took {time() - t} s.")

        # TODO MM on adding asyn-trainer change api to get timeline
        t = time()
        self._report_train_metrics(
            model=self.global_model(),
            timeline=timeline,
            metric_reporter=metric_reporter,
        )
        self._report_post_aggregation_train_metrics(
            clients=agg_metric_clients,
            model=self.global_model(),
            timeline=timeline,
            metric_reporter=metric_reporter,
        )
        self._calc_post_epoch_communication_metrics(
            timeline,
            metric_reporter,
        )
        self.logger.info(f"Aggregate round reporting took {time() - t} s.")

    def _choose_clients_for_post_aggregation_metrics(
        self,
        train_clients: Iterable[Client],
        num_total_users: int,
        users_per_round: int,
    ) -> Iterable[Client]:
        """
        Chooses clients for the post-aggregation training metrics.
        Depending on config parameters, either return the round's
        training clients or new randomly drawn clients.
        """
        # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
        if self.cfg.use_train_clients_for_aggregation_metrics:
            return train_clients

        # For the post-aggregation metrics, evaluate on new users
        agg_metric_client_idcs = torch.multinomial(
            torch.ones(num_total_users, dtype=torch.float),
            users_per_round,
            replacement=False,
        ).tolist()

        agg_metric_clients = [
            self.create_or_get_client_for_data(i, self.data_provider)
            for i in agg_metric_client_idcs
        ]
        return agg_metric_clients

    def calc_post_aggregation_train_metrics(
        self,
        clients: Iterable[Client],
        model: IFLModel,
        timeline: Timeline,
        metric_reporter: Optional[IFLMetricsReporter],
    ) -> List[Metric]:
        """
        Calculates post-server aggregation metrics.
        """
        # TODO do multiprocessing
        # lines below internally change the state of metric reporter,
        # no need to create extra reportable metrics.
        [c.eval(model=model, metric_reporter=metric_reporter) for c in clients]

        metrics = []
        if self.is_user_level_dp:
            user_eps = self.server.privacy_budget.epsilon  # pyre-fixme
            metrics.append(Metric("user level dp (eps)", user_eps))
        if self.is_sample_level_dp:
            # calculate sample level dp privacy loss statistics.
            all_client_eps = torch.Tensor(
                [c.privacy_budget.epsilon for c in clients]  # pyre-fixme
            )
            mean_client_eps = all_client_eps.mean()
            max_client_eps = all_client_eps.max()
            min_client_eps = all_client_eps.min()
            p50_client_eps = torch.median(all_client_eps)
            sample_dp_metrics: List[Metric] = Metric.from_args(
                mean=mean_client_eps,
                min=min_client_eps,
                max=max_client_eps,
                median=p50_client_eps,
            )
            metrics.append(Metric("sample level dp (eps)", sample_dp_metrics))

        return metrics

    def calc_post_epoch_client_metrics(
        self,
        client_models: Dict[Client, IFLModel],
        round_timeline: Timeline,
        metric_reporter: IFLMetricsReporter,
    ) -> List[List[Metric]]:
        """
        Calculates client-side metrics on the overall evaluation set
        """
        client_metrics = []
        if metric_reporter is not None:
            for client, model in tqdm(client_models.items()):
                metric_reporter.reset()
                client.eval(
                    model=model,
                    dataset=self.data_provider.eval_data(),
                    metric_reporter=metric_reporter,
                )
                # pyre-fixme[16]: `IFLMetricsReporter` has no attribute
                #  `compute_scores`.
                score = metric_reporter.compute_scores()
                client_metrics.append(Metric.from_dict(score))

        return client_metrics

    def _report_post_aggregation_train_metrics(
        self,
        clients: Iterable[Client],
        model: IFLModel,
        timeline: Timeline,
        metric_reporter: Optional[IFLMetricsReporter],
    ):

        if (
            metric_reporter is not None
            # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
            and self.cfg.report_train_metrics
            and self.cfg.report_train_metrics_after_aggregation
            and timeline.tick(1.0 / self.cfg.train_metrics_reported_per_epoch)
        ):
            print(f"reporting {timeline} for aggregation")
            metrics = self.calc_post_aggregation_train_metrics(
                clients, model, timeline, metric_reporter
            )

            metric_reporter.report_metrics(
                model=model,
                reset=True,
                stage=TrainingStage.AGGREGATION,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
                extra_metrics=metrics,
            )

    def _validate_users_per_round(
        self, users_per_round_on_worker: int, num_users_on_worker: int
    ):
        assert users_per_round_on_worker <= num_users_on_worker, (
            "Users per round is greater than number of users in data provider for the worker."
            "If you are using paged dataloader, increase your num_users_per_page >> users_per_round"
        )

    def _report_post_epoch_client_metrics(
        self,
        timeline: Timeline,
        metric_reporter: Optional[IFLMetricsReporter],
    ):
        if (
            metric_reporter is not None
            # pyre-fixme[16]: `SyncTrainer` has no attribute `cfg`.
            and self.cfg.report_client_metrics
            and self.cfg.report_client_metrics_after_epoch
            and (timeline.epoch % self.cfg.client_metrics_reported_per_epoch == 0)
        ):
            client_models = {
                client: client.last_updated_model for client in self.clients.values()
            }
            client_scores = self.calc_post_epoch_client_metrics(
                client_models, timeline, metric_reporter
            )

            # Find stats over the client_metrics (mean, min, max, median, std)
            client_stats_trackers = {}
            score_names = [metric.name for metric in next(iter(client_scores))]
            for score_name in score_names:
                client_stats_trackers[score_name] = RandomVariableStatsTracker(
                    tracks_quantiles=True
                )
            for client_metric_list in client_scores:
                for client_metric in client_metric_list:
                    client_stats_trackers[client_metric.name].update(
                        client_metric.value
                    )

            reportable_client_metrics = []
            for score_name in score_names:
                for stat_name, stat_key in [
                    ("Mean", "mean_val"),
                    ("Median", "median_val"),
                    ("Upper Quartile", "upper_quartile_val"),
                    ("Lower Quartile", "lower_quartile_val"),
                    ("Min", "min_val"),
                    ("Max", "max_val"),
                    ("Standard Deviation", "standard_deviation_val"),
                    ("Num Samples", "num_samples"),
                ]:
                    score = client_stats_trackers[score_name].__getattribute__(stat_key)
                    reportable_client_metrics.append(Metric(stat_name, score))

            metric_reporter.report_metrics(
                model=None,
                reset=True,
                stage=TrainingStage.PER_CLIENT_EVAL,
                timeline=timeline,
                epoch=timeline.global_round_num(),  # for legacy
                print_to_channels=True,
                extra_metrics=reportable_client_metrics,
            )

    @staticmethod
    def rounds_in_one_epoch(num_total_users: int, users_per_round: int) -> int:
        return math.ceil(num_total_users / users_per_round)


def force_print(is_distributed: bool, *args, **kwargs) -> None:
    if is_distributed:
        try:
            device_info = f" [device:{torch.cuda.current_device()}]"
            # pyre-fixme[28]: Unexpected keyword argument `force`.
            print(*args, device_info, **kwargs, force=True)
        except TypeError:
            pass
    else:
        print(*args, **kwargs)


@dataclass
class SyncTrainerConfig(FLTrainerConfig):
    _target_: str = fullclassname(SyncTrainer)
    server: SyncServerConfig = SyncServerConfig()
    users_per_round: int = 10
    # overselect users_per_round / dropout_rate users, only use first
    # users_per_round updates
    dropout_rate: float = 1.0
    report_train_metrics_after_aggregation: bool = False
    report_client_metrics_after_epoch: bool = False
    # Whether client metrics on eval data should be computed and reported.
    report_client_metrics: bool = False
    # how many times per epoch should we report client metrics
    # numbers greater than 1 help with plotting more precise training curves
    client_metrics_reported_per_epoch: int = 1



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Optional


class ProcessState:
    _instance: Optional["ProcessState"] = None

    @staticmethod
    def getInstance(**kwargs):
        """kwargs should specify:
        rank: int, workflow_name: Optional[str], chronos_id: Optional[int]
        """
        if ProcessState._instance is None:
            ProcessState(**kwargs)
        return ProcessState._instance

    def __init__(self, rank: int):
        """
        Virtually private constructor.
        Handles logic for Singleton pattern.
        """
        self._rank = rank
        if ProcessState._instance is not None:
            raise RuntimeError(
                "ProcessState is a singleton. Cannot instantiate multiple times!"
            )
        else:
            ProcessState._instance = self

    @property
    def rank(self):
        return self._rank

    @rank.setter
    def rank(self, value):
        if self._rank is not None:
            raise RuntimeError("Shouldn't change 'rank' after initialized")
        self._rank = value



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.utils.async_trainer.device_state import DeviceState, TrainingSchedule
from libfb.py import testutil


class DeviceStateUtilTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_next_event_time_update(self):
        """Check that next_event_time in DeviceState is updated correctly depending
        on its state
        """
        training_schedule = TrainingSchedule(
            creation_time=0, start_time=12, end_time=16
        )
        device_state = DeviceState(training_schedule)
        self.assertEqual(device_state.next_event_time(), training_schedule.start_time)
        device_state.training_started()
        self.assertEqual(device_state.next_event_time(), training_schedule.end_time)
        device_state.training_ended()
        self.assertEqual(device_state.next_event_time(), training_schedule.end_time)

    def test_device_next_event_time_comparison(self):
        """Check whether comparison operator for DeviceState acts as expected"""
        # create two devices, 1 & 2
        # device1 has current, training_start and training_end times
        # that are slightly before device2's respective times
        # 1. verify initial ordering
        # 2. advance training for device1 followed by device2,
        # verify ordering is as expected in every step
        training_schedule1 = TrainingSchedule(
            creation_time=0, start_time=12, end_time=16
        )
        device_state_1 = DeviceState(training_schedule1)
        training_schedule2 = TrainingSchedule(
            creation_time=0, start_time=12.1, end_time=16.1
        )
        device_state_2 = DeviceState(training_schedule2)
        self.assertLess(device_state_1, device_state_2)

        device_state_1.training_started()
        self.assertGreater(device_state_1, device_state_2)

        device_state_2.training_started()
        self.assertLess(device_state_1, device_state_2)

        device_state_1.training_ended()
        self.assertLess(device_state_1, device_state_2)

        device_state_2.training_ended()
        self.assertLess(device_state_1, device_state_2)

    def test_device_next_event_time_comparison_equality(self):
        """Check whether comparison operator for DeviceState acts as expected
        when next_event_time is equal
        """
        # create two devices, 1 & 2
        # device1.training_start = device2.current_time
        # device1.training_end = device2.training_schedule
        # 1. verify initial ordering
        # 2. advance device training
        # verify ordering is as expected in every step: when next_event_time is
        # equal, device that is 'further along' in training is 'lesser'
        training_schedule1 = TrainingSchedule(
            creation_time=0, start_time=12, end_time=16
        )
        device_state_1 = DeviceState(training_schedule1)
        training_schedule2 = TrainingSchedule(
            creation_time=0, start_time=16, end_time=20
        )
        device_state_2 = DeviceState(training_schedule2)
        self.assertLess(device_state_1, device_state_2)

        device_state_1.training_started()  # relevant time for both is 12
        # device that is further along in training is 'Less'
        self.assertLess(device_state_1, device_state_2)
        # verify that ordering is independent of which device_state is the
        # first parameter
        self.assertGreater(device_state_2, device_state_1)

        device_state_2.training_started()
        self.assertLess(device_state_1, device_state_2)

        device_state_1.training_ended()
        self.assertLess(device_state_1, device_state_2)

        device_state_2.training_ended()
        self.assertLess(device_state_1, device_state_2)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import Type

import pytest
from flsim.channels.communication_stats import (
    ChannelDirection,
)
from flsim.channels.message import Message
from flsim.channels.product_quantization_channel import (
    ProductQuantizationChannelConfig,
    ProductQuantizationChannel,
)
from flsim.common.pytest_helper import assertEqual, assertIsInstance
from flsim.tests import utils
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate


class TestProductQuantizationChannel:
    @pytest.mark.parametrize(
        "config",
        [
            {
                "_target_": ProductQuantizationChannelConfig._target_,
                "max_num_centroids": 256,
                "num_codebooks": 1,
                "max_block_size": 9,
            }
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ProductQuantizationChannel],
    )
    def test_pq_instantiation(self, config: Type, expected_type: Type) -> None:
        """
        Tests instantiation of the scalar quantization channel.
        """

        # test instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

    @pytest.mark.parametrize(
        "config",
        [
            {
                "_target_": ProductQuantizationChannelConfig._target_,
                "max_num_centroids": 4,
                "num_codebooks": 1,
                "max_block_size": 2,
            },
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ProductQuantizationChannel],
    )
    def test_pq_server_to_client(self, config: Type, expected_type: Type) -> None:
        """
        Tests server to client transmission of the message. Models
        before and after transmission should be identical since we
        only product quantize in the client to server direction.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)

        # test server -> client, models should be strictly identical
        message = Message(download_model)
        message = channel.server_to_client(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), download_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [
            {
                "_target_": ProductQuantizationChannelConfig._target_,
                "max_num_centroids": 2,
                "num_codebooks": 1,
                "max_block_size": 1,
            },
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ProductQuantizationChannel],
    )
    def test_pq_client_to_server(self, config: Type, expected_type: Type) -> None:
        """
        Tests client to server transmission of the message. Models
        before and after transmission should be identical since we quantize
        with 2 codebooks of 1 centroid each with constnt weights.

        Notes:
            - Product Quantization itself is tested in `test_product_quantization`.
        """

        # instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

        # create dummy model
        two_fc = utils.TwoFC()
        two_fc.fc1.weight.data[:, 0] = 1
        two_fc.fc1.weight.data[:, 1] = 2
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, models should be almost equal due to int8 emulation
        message = Message(upload_model)
        message = channel.client_to_server(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), upload_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [
            {
                "_target_": ProductQuantizationChannelConfig._target_,
                "max_num_centroids": 2,
                "num_codebooks": 1,
                "max_block_size": 1,
                "min_numel_to_quantize": 10,
                "report_communication_metrics": True,
            },
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [ProductQuantizationChannel],
    )
    def test_pq_stats(self, config: Type, expected_type: Type) -> None:
        """
        Tests stats measurement. We manually compute bytes sent from client
        to the server and verify that it matches the bytes computed by the
        channel. Here, we quantize with 1 bit per weight of assignment and
        need to store one centroids of one byte per layer.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # client -> server
        message = Message(upload_model)
        message = channel.client_to_server(message)

        # test communication stats measurements (here per_tensor int8 quantization)
        stats = channel.stats_collector.get_channel_stats()
        client_to_server_bytes = stats[ChannelDirection.CLIENT_TO_SERVER].mean()

        # biases are not quantized
        n_biases = 6
        bias_size_bytes = n_biases * channel.BYTES_PER_FP32

        # layer 2 with 5 elements is not quantized since min_numel_to_quantize = 10
        n_weights_layer_1 = 10
        n_weights_layer_2 = 5

        # weights of layer 1 are quantized with 1 bit per weight of assignments
        weight_size_bytes = n_weights_layer_1 / 8.0

        # weights of layer 2 are not quantized and stored in `fp32`
        weight_size_bytes += n_weights_layer_2 * channel.BYTES_PER_FP32

        # storing codebook, two `fp32` centroid for the first layer
        centroids_bytes = 2 * channel.BYTES_PER_FP32

        compressed_bytes = bias_size_bytes + weight_size_bytes + centroids_bytes

        assertEqual(client_to_server_bytes, compressed_bytes)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""
This file defines the concept of a base client for a
federated learning setting. Also defines basic config,
for an FL client.

Note:
    This is just a base class and needs to be overridden
    for different use cases.
"""
from __future__ import annotations

import logging
import random
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Iterable

import torch
from flsim.channels.base_channel import IdentityChannel
from flsim.channels.message import Message
from flsim.common.logger import Logger
from flsim.common.timeout_simulator import (
    NeverTimeOutSimulator,
    NeverTimeOutSimulatorConfig,
    TimeOutSimulator,
)
from flsim.data.data_provider import IFLUserData
from flsim.interfaces.metrics_reporter import IFLMetricsReporter
from flsim.interfaces.model import IFLModel
from flsim.optimizers.local_optimizers import (
    LocalOptimizerConfig,
    LocalOptimizerSGDConfig,
)
from flsim.optimizers.optimizer_scheduler import (
    OptimizerScheduler,
    OptimizerSchedulerConfig,
    ConstantLRSchedulerConfig,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.cuda import ICudaStateManager, DEFAULT_CUDA_MANAGER
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate
from libfb.py.pyre import none_throws
from omegaconf import OmegaConf


class Client:
    logger: logging.Logger = Logger.get_logger(__name__)

    def __init__(
        self,
        *,
        dataset: IFLUserData,
        channel: Optional[IdentityChannel] = None,
        timeout_simulator: Optional[TimeOutSimulator] = None,
        store_last_updated_model: Optional[bool] = False,
        cuda_manager: ICudaStateManager = DEFAULT_CUDA_MANAGER,
        name: Optional[str] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=ClientConfig,
            **kwargs,
        )

        self.dataset = dataset
        self.cuda_state_manager = cuda_manager
        self.channel = channel or IdentityChannel()
        self.timeout_simulator = timeout_simulator or NeverTimeOutSimulator(
            **OmegaConf.structured(NeverTimeOutSimulatorConfig())
        )
        self.store_last_updated_model = store_last_updated_model
        self.name = name or "unnamed_client"

        # base lr needs to match LR in optimizer config, overwrite it
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        self.cfg.lr_scheduler.base_lr = self.cfg.optimizer.lr
        self.per_example_training_time = (
            self.timeout_simulator.simulate_per_example_training_time()
        )
        self.ref_model = None
        self.num_samples = 0
        self.times_selected = 0
        self._tracked = {}
        self._last_updated_model = None
        self.logger.setLevel(logging.INFO)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.optimizer, "_target_"):
            cfg.optimizer = LocalOptimizerSGDConfig()
        if OmegaConf.is_missing(cfg.lr_scheduler, "_target_"):
            cfg.lr_scheduler = ConstantLRSchedulerConfig()

    @property
    def seed(self) -> Optional[int]:
        """if should set random_seed or not."""
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        return self.cfg.random_seed

    @property
    def model_deltas(self) -> List[IFLModel]:
        """
        return the stored deltas for all rounds that
        this user was selected.
        """
        return [self._tracked[s]["delta"] for s in range(self.times_selected)]

    @property
    def last_updated_model(self):
        """
        return the most recent model on the client
        """
        return self._last_updated_model

    @property
    def optimizers(self) -> List[Any]:
        """Look at {self.model}"""
        return [self._tracked[s]["optimizer"] for s in range(self.times_selected)]

    @property
    def weights(self) -> List[float]:
        """Look at {self.model}"""
        return [self._tracked[s]["weight"] for s in range(self.times_selected)]

    def generate_local_update(
        self, model: IFLModel, metric_reporter: Optional[IFLMetricsReporter] = None
    ) -> Tuple[IFLModel, float]:
        r"""
        wrapper around all functions called on a client for generating an
        updated local model.

        Note:
        -----
        Only pass a ``metric_reporter`` if reporting is needed, i.e.
        report_metrics will be called ont the reporter, o.w. reports will be
        accumulated in memory.
        """
        # 1. pass through channel, set initial state
        updated_model = self.receive_through_channel(model)
        # 2. set up model and optimizer in the client
        updated_model, optim, optim_scheduler = self.prepare_for_training(updated_model)
        # 3. kick off training on client
        updated_model, weight = self.train(
            updated_model, optim, optim_scheduler, metric_reporter
        )
        # 4. Store updated model if being tracked
        if self.store_last_updated_model:
            self._last_updated_model = deepcopy(updated_model)
        # 5. compute delta
        delta = self.compute_delta(
            before=model, after=updated_model, model_to_save=updated_model
        )
        # 6. track state of the client
        self.track(delta=delta, weight=weight, optimizer=optim)
        return delta, weight

    def compute_delta(
        self, before: IFLModel, after: IFLModel, model_to_save: IFLModel
    ) -> IFLModel:
        """
        Computes the delta between the before training and after training model
        """
        FLModelParamUtils.subtract_model(
            minuend=before.fl_get_module(),
            subtrahend=after.fl_get_module(),
            difference=model_to_save.fl_get_module(),
        )
        return model_to_save

    def receive_through_channel(self, model: IFLModel) -> IFLModel:
        """
        Receives a reference to a state (refered to a model state_dict)
        over the channel. Any channel effect is applied as part of this
        receive function.
        """
        # keep a reference to global model
        self.ref_model = model

        # need to deepcopy the model because it's a reference to the global model
        # modifying model will moidify the global model
        message = self.channel.server_to_client(Message(model=deepcopy(model)))

        return message.model

    def prepare_for_training(
        self, model: IFLModel
    ) -> Tuple[IFLModel, torch.optim.Optimizer, OptimizerScheduler]:
        """
        1- instansiate a model with the given initial state
        2- create an optimizer
        """
        # inform cuda_state_manager that we're about to train a model
        # it may move model to GPU
        self.cuda_state_manager.before_train_or_eval(model)
        # put model in train mode
        model.fl_get_module().train()
        # create optimizer
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        optimizer = instantiate(self.cfg.optimizer, model=model.fl_get_module())
        optimizer_scheduler = instantiate(self.cfg.lr_scheduler, optimizer=optimizer)
        return model, optimizer, optimizer_scheduler

    def get_total_training_time(self) -> float:
        return self.timeout_simulator.simulate_training_time(
            self.per_example_training_time, self.dataset.num_examples()
        )

    def stop_training(self, num_examples_processed) -> bool:
        training_time = self.timeout_simulator.simulate_training_time(
            self.per_example_training_time, num_examples_processed
        )
        return self.timeout_simulator.user_timeout(training_time)

    def train(
        self,
        model: IFLModel,
        optimizer: Any,
        optimizer_scheduler: OptimizerScheduler,
        metric_reporter: Optional[IFLMetricsReporter] = None,
    ) -> Tuple[IFLModel, float]:
        total_samples = 0
        # NOTE currently weight = total_sampls, this might be a bad strategy
        # plus there are privcay implications that must be taken into account.
        num_examples_processed = 0  # number of examples processed during training
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        for epoch in range(self.cfg.epochs):
            if self.stop_training(num_examples_processed):
                break

            # if user has too many examples and times-out, we want to process
            # different portion of the dataset each time
            dataset = list(iter(self.dataset))
            if self.cfg.shuffle_batch_order:
                random.shuffle(dataset)
            for batch in dataset:
                # TODO use an independent random generator
                if self.seed is not None:
                    torch.manual_seed(self.seed)

                sample_count = self._batch_train(
                    model=model,
                    optimizer=optimizer,
                    training_batch=batch,
                    epoch=epoch,
                    metric_reporter=metric_reporter,
                    optimizer_scheduler=optimizer_scheduler,
                )
                self.post_batch_train(epoch, model, sample_count, optimizer)
                total_samples += 0 if epoch else sample_count
                num_examples_processed += sample_count
                # stop training depending on time-out condition
                if self.stop_training(num_examples_processed):
                    break
        # tell cuda manager we're done with training
        # cuda manager may move model out of GPU memory if needed
        self.cuda_state_manager.after_train_or_eval(model)
        self.logger.debug(
            f"Processed {num_examples_processed} of {self.dataset.num_examples()}"
        )
        self.post_train(model, total_samples, optimizer)
        # if training stops early, used partial training weight
        example_weight = min([num_examples_processed, total_samples])
        return model, float(example_weight)

    def post_train(self, model: IFLModel, total_samples: int, optimizer: Any):
        pass

    def post_batch_train(
        self, epoch: int, model: IFLModel, sample_count: int, optimizer: Any
    ):
        pass

    def track(self, delta: IFLModel, weight: float, optimizer: Any):
        """Tracks metric when the client is selected multiple times"""
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        if self.cfg.store_models_and_optimizers:
            self._tracked[self.times_selected] = {
                "delta": deepcopy(delta),
                "weight": weight,
                "optimizer": optimizer,
            }
        self.times_selected += 1

    def eval(
        self,
        model: Optional[IFLModel] = None,
        dataset: Optional[Iterable] = None,
        metric_reporter: Optional[IFLMetricsReporter] = None,
    ):
        """
        Evaluate the given `model` with the given `dataset`. `model` defaults
        to the current global_model if nothing is provided, and `dataset`
        defaults to client's dataset.
        """
        # Note here we play a trick, model and dataset provided are not
        # passed through a channel, in ptactice this needs to be done if
        # either the model or dataset is not what is currently being held
        # locally at the client.
        model = model or self.ref_model
        data = dataset or self.dataset
        self.cuda_state_manager.before_train_or_eval(model)
        with torch.no_grad():
            if self.seed is not None:
                torch.manual_seed(self.seed)

            model.fl_get_module().eval()
            for batch in data:
                batch_metrics = model.get_eval_metrics(batch)
                if metric_reporter is not None:
                    # TODO MM make sure metric reporter is multi-process safe.
                    metric_reporter.add_batch_metrics(batch_metrics)
        model.fl_get_module().train()
        self.cuda_state_manager.after_train_or_eval(model)

    def _batch_train(
        self,
        model,
        optimizer,
        training_batch,
        epoch,
        metric_reporter,
        optimizer_scheduler,
    ) -> int:
        """Trainer for NewDocModel based FL Tasks
        Run a single iteration of minibatch-gradient descent on a single user.
        Compatible with the new tasks in which model is reponsible to
        arrange its inputs, targets and context.
        Return number of examples in the batch.
        """
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                "L1-norm of Parameters of initial state:",
                sum(p.abs().sum() for p in self.ref_model.fl_get_module().parameters()),
            )
            self.logger.debug(
                "L1-norm of Parameters before step:",
                sum(p.abs().sum() for p in model.fl_get_module().parameters()),
            )

        optimizer.zero_grad()
        batch_metrics = model.fl_forward(training_batch)
        loss = batch_metrics.loss

        loss.backward()
        # pyre-fixme[16]: `Client` has no attribute `cfg`.
        if self.cfg.max_clip_norm_normalized is not None:
            max_norm = none_throws(self.cfg.max_clip_norm_normalized)
            FLModelParamUtils.clip_gradients(
                max_normalized_l2_norm=max_norm, model=model.fl_get_module()
            )

        num_examples = batch_metrics.num_examples
        # adjust lr and take a step
        optimizer_scheduler.step(batch_metrics, model, training_batch, epoch)
        optimizer.step()

        if metric_reporter is not None:
            metric_reporter.add_batch_metrics(batch_metrics)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                "L1-norm of Parameters after step:",
                sum(p.abs().sum() for p in model.fl_get_module().parameters()),
            )
            l2_norm_raw = FLModelParamUtils.get_gradient_l2_norm_raw(
                model.fl_get_module()
            )
            l2_norm_normalized = FLModelParamUtils.get_gradient_l2_norm_normalized(
                model.fl_get_module()
            )
            self.logger.debug(
                f"Train Loss:{loss},"
                f"GradNorm:{l2_norm_raw},"
                f"NormalizedGradNorm:{l2_norm_normalized},",
                f"NumExamples:{num_examples}",
            )
        return num_examples


@dataclass
class ClientConfig:
    _target_: str = fullclassname(Client)
    _recursive_: bool = False
    epochs: int = 1  # No. of epochs for local training
    optimizer: LocalOptimizerConfig = LocalOptimizerConfig()
    lr_scheduler: OptimizerSchedulerConfig = OptimizerSchedulerConfig()
    max_clip_norm_normalized: Optional[float] = None  # gradient clip value
    only_federated_params: bool = True  # flag to only use certain params
    random_seed: Optional[int] = None  # random seed for deterministic response
    shuffle_batch_order: bool = False  # shuffle the ordering of batches
    store_models_and_optimizers: bool = False  # name clear
    track_multiple_selection: bool = False  # track if client appears in 2+ rounds.



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.channels.sketch_channel import SketchChannelConfig
from flsim.common.pytest_helper import assertIsInstance, assertEqual
from flsim.optimizers.sketch_aggregator import SketchAggregator, SketchAggregatorConfig
from flsim.tests import utils
from hydra.utils import instantiate


def test_sketch_aggregator() -> None:
    """
    Tests Sketch Aggregator and make sure its init, init_round, collect_update,
    and step are all correct.
    """

    # init
    channel = instantiate(SketchChannelConfig(independence=4))
    model = utils.SampleNet(utils.TwoFC())
    global_param = 1.0
    model.fl_get_module().fill_all(global_param)

    aggregator = instantiate(
        SketchAggregatorConfig(top_k_ratio=0.5),
        global_model=model,
        channel=channel,
    )
    assertIsInstance(aggregator, SketchAggregator)
    assertEqual(0.5, aggregator.cfg.top_k_ratio)
    assertEqual(4, aggregator.grad_cs.independence)
    assert torch.allclose(aggregator.grad_cs.buckets, torch.tensor([0.0]))
    assert torch.allclose(aggregator.error_cs.buckets, torch.tensor([0.0]))

    # init_round
    aggregator.init_round()
    assert torch.allclose(aggregator.reducer.round_cs.buckets, torch.tensor([0.0]))

    # collect_client_update
    model = utils.SampleNet(utils.TwoFC())
    model.fl_get_module().fill_all(0.6)
    aggregator.collect_client_update(model, 2.0)

    model = utils.SampleNet(utils.TwoFC())
    model.fl_get_module().fill_all(0.3)
    aggregator.collect_client_update(model, 1.0)

    expected_value1 = 0.6 * 2.0 + 0.3 * 1.0
    total_weight = 2.0 + 1.0

    params = aggregator.reducer.round_cs.unsketch_model()
    for _, param in params.items():
        assert torch.allclose(param, torch.tensor([expected_value1]))

    # step
    aggregator.step()
    ones = 0
    correct = 0
    for _, param in aggregator.global_model.fl_get_module().named_parameters():
        ones += torch.sum(param == 1)
        correct += torch.sum(
            param == global_param - expected_value1 / total_weight * aggregator.cfg.lr
        )

    assertEqual(correct, aggregator.top_k)
    assertEqual(ones, aggregator.grad_cs.n - aggregator.top_k)

    # init round
    aggregator.init_round()

    # collect_client_update again
    model = utils.SampleNet(utils.TwoFC())
    model.fl_get_module().fill_all(1.2)
    aggregator.collect_client_update(model, 1.0)

    model = utils.SampleNet(utils.TwoFC())
    model.fl_get_module().fill_all(0.9)
    aggregator.collect_client_update(model, 2.0)

    expected_value2 = 0.9 * 2.0 + 1.2 * 1.0
    total_weight = 2.0 + 1.0

    params = aggregator.reducer.round_cs.unsketch_model()
    for _, param in params.items():
        assert torch.allclose(param, torch.tensor([expected_value2]))

    # step again
    expected_grad = (
        aggregator.cfg.momentum * expected_value1 / total_weight
        + expected_value2 / total_weight
    )

    aggregator.step()
    assert torch.allclose(
        aggregator.grad_cs.query(torch.arange(0, 21)),
        torch.tensor([expected_grad]),
    )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from typing import OrderedDict
from typing import Type

import pytest
import torch
from flsim.channels.base_channel import Message
from flsim.channels.communication_stats import (
    ChannelDirection,
)
from flsim.channels.sparse_mask_channel import (
    SparseMaskChannel,
    SparseMaskChannelConfig,
)
from flsim.common.pytest_helper import assertEqual, assertIsInstance, assertAlmostEqual
from flsim.tests import utils
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.tests.helpers.test_models import FCModel
from hydra.utils import instantiate
from torch.nn.utils import prune


class TestSparseMaskChannel:
    @classmethod
    def calc_model_sparsity(cls, state_dict: OrderedDict):
        """
        Calculates model sparsity (fraction of zeroed weights in state_dict).
        """
        non_zero = 0
        tot = 1e-6
        for _, param in state_dict.items():
            non_zero += torch.count_nonzero(param).item()
            tot += float(param.numel())
        return 1.0 - non_zero / (tot + 1e-6)

    def test_sparse_model_size(self) -> None:
        model = FCModel()
        # Prune model to a quarter of its size
        params_to_prune = [
            (model.fc1, "weight"),
            (model.fc1, "bias"),
            (model.fc2, "weight"),
            (model.fc2, "bias"),
            (model.fc3, "weight"),
            (model.fc3, "bias"),
        ]
        prune.global_unstructured(
            params_to_prune,
            pruning_method=prune.L1Unstructured,
            amount=0.75,
        )
        for module, name in params_to_prune:
            prune.remove(module, name)
        sparsity = self.calc_model_sparsity(model.state_dict())
        assertAlmostEqual(
            0.75,
            sparsity,
            delta=0.02,  # Accounts for 2 percentage points difference
        )

    @pytest.mark.parametrize(
        "config",
        [
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="topk",
            ),
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="random",
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SparseMaskChannel],
    )
    def test_random_mask_instantiation(self, config: Type, expected_type: Type) -> None:
        """
        Tests instantiation of the random mask channel.
        """

        # test instantiation
        channel = instantiate(config)
        assertIsInstance(channel, expected_type)

    @pytest.mark.parametrize(
        "config",
        [
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="topk",
            ),
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="random",
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SparseMaskChannel],
    )
    def test_random_mask_server_to_client(
        self, config: Type, expected_type: Type
    ) -> None:
        """
        Tests server to client transmission of the message. Models
        before and after transmission should be identical since
        we random mask only on the client to server direction.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        download_model = deepcopy(base_model)

        # test server -> client, models should be strictly identical
        message = Message(download_model)
        message = channel.server_to_client(message)

        mismatched = FLModelParamUtils.get_mismatched_param(
            [base_model.fl_get_module(), download_model.fl_get_module()]
        )
        assertEqual(mismatched, "", mismatched)

    @pytest.mark.parametrize(
        "config",
        [
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="topk",
            ),
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="random",
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SparseMaskChannel],
    )
    def test_sparse_mask_client_to_server(
        self, config: Type, expected_type: Type
    ) -> None:
        """
        Tests client to server transmission of the message. Model
        after transmission should have the right sparsity ratio.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, check for sparsity ratio
        message = Message(upload_model)
        message = channel.client_to_server(message)

        sparsity = self.calc_model_sparsity(message.model_state_dict)

        # sparsity ratio should be approximately proportion_of_zero_weights
        # approximately since we round the number of parameters to prune
        # to an integer, see sparse_mask_channel.py
        assertAlmostEqual(channel.cfg.proportion_of_zero_weights, sparsity, delta=0.05)

    @pytest.mark.parametrize(
        "config",
        [
            SparseMaskChannelConfig(
                proportion_of_zero_weights=0.6,
                sparsity_method="topk",
            ),
        ],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SparseMaskChannel],
    )
    def test_topk_mask_sparsity(self, config: Type, expected_type: Type) -> None:
        """
        Tests that TopK compression has worked ie. the smallest client updates
        are  masked out.
        """

        # instantiation
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # test client -> server, check for topk sparse mask
        message = Message(upload_model)
        message = channel.client_to_server(message)

        for name, p in base_model.fl_get_module().named_parameters():
            flattened_params = p.flatten().abs()
            sparse_indices = flattened_params.abs().argsort()[
                : int(config.proportion_of_zero_weights * flattened_params.numel())
            ]
            flattened_message_params = torch.cat(
                [torch.flatten(p) for p in message.model_state_dict[name]]
            ).flatten()
            assertEqual(flattened_message_params[sparse_indices].sum(), 0.0)

    @pytest.mark.parametrize("sparsity_method", ["topk", "random"])
    @pytest.mark.parametrize(
        "compressed_size_measurement",
        ["bitmask", "coo"],
    )
    @pytest.mark.parametrize(
        "expected_type",
        [SparseMaskChannel],
    )
    def test_sparse_mask_stats(
        self,
        sparsity_method: str,
        compressed_size_measurement: str,
        expected_type: Type,
    ) -> None:
        """
        Tests stats measurement. We assume that the sparse tensor
        is stored in COO format and manually compute the number
        of bytes sent from client to server to check that it
        matches that the channel computes.
        """

        # instantiation
        config = SparseMaskChannelConfig(
            proportion_of_zero_weights=0.6,
            report_communication_metrics=True,
            sparsity_method=sparsity_method,
            compressed_size_measurement=compressed_size_measurement,
        )
        channel = instantiate(config)

        # create dummy model
        two_fc = utils.TwoFC()
        base_model = utils.SampleNet(two_fc)
        upload_model = deepcopy(base_model)

        # client -> server
        message = Message(upload_model)
        message = channel.client_to_server(message)

        # test communication stats measurements
        stats = channel.stats_collector.get_channel_stats()
        client_to_server_bytes = stats[ChannelDirection.CLIENT_TO_SERVER].mean()

        # compute sizes
        n_weights = sum([p.numel() for p in two_fc.parameters() if p.ndim == 2])
        n_biases = sum([p.numel() for p in two_fc.parameters() if p.ndim == 1])
        non_zero_weights = n_weights - int(
            n_weights * channel.cfg.proportion_of_zero_weights
        )
        non_zero_biases = n_biases - int(
            n_biases * channel.cfg.proportion_of_zero_weights
        )
        n_dim_weights = 2
        n_dim_biases = 1
        true_size_bytes_weights = (
            # size of the index
            (
                non_zero_weights * SparseMaskChannel.BYTES_PER_INT64 * n_dim_weights
                if compressed_size_measurement == "coo"
                else SparseMaskChannel.BYTES_PER_BIT * n_weights
            )
            # size of values
            + non_zero_weights * SparseMaskChannel.BYTES_PER_FP32
        )
        true_size_bytes_biases = (
            # size of the index
            (
                non_zero_biases * SparseMaskChannel.BYTES_PER_INT64 * n_dim_biases
                if compressed_size_measurement == "coo"
                else SparseMaskChannel.BYTES_PER_BIT * n_biases
            )
            # size of values
            + non_zero_biases * SparseMaskChannel.BYTES_PER_FP32
        )

        # size of the values
        true_size_bytes = true_size_bytes_weights + true_size_bytes_biases

        assertEqual(client_to_server_bytes, true_size_bytes)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional

from flsim.channels.sketch_channel import SketchChannel
from flsim.common.logger import Logger
from flsim.interfaces.model import IFLModel
from flsim.optimizers.sync_aggregators import SyncAggregator, SyncAggregatorConfig
from flsim.reducers.base_round_reducer import IFLRoundReducer
from flsim.reducers.sketch_round_reducer import (
    SketchRoundReducer,
    SketchRoundReducerConfig,
)
from flsim.utils.config_utils import fullclassname, init_self_cfg
from flsim.utils.count_sketch import CountSketch, clone_count_sketch
from flsim.utils.fl.common import FLModelParamUtils
from omegaconf import OmegaConf


class SketchAggregator(SyncAggregator):
    """
    Implements FetchSGD
    """

    logger: logging.Logger = Logger.get_logger("SketchAggregator")

    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: SketchChannel,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SketchAggregatorConfig,
            **kwargs,
        )

        super().__init__(global_model=global_model, channel=channel, **kwargs)

        device = next(global_model.fl_get_module().parameters()).device
        self.grad_cs = CountSketch(
            # pyre-fixme[16]: `SketchChannel` has no attribute `cfg`.
            width=channel.cfg.num_col,
            depth=channel.cfg.num_hash,
            prime=channel.cfg.prime,
            independence=channel.cfg.independence,
            h=channel.h,
            g=channel.g,
            device=device,
        )
        self.error_cs = CountSketch(
            width=channel.cfg.num_col,
            depth=channel.cfg.num_hash,
            prime=channel.cfg.prime,
            independence=channel.cfg.independence,
            h=channel.h,
            g=channel.g,
            device=device,
        )
        self.grad_cs.set_params(global_model.fl_get_module().state_dict())
        self.error_cs.set_params(global_model.fl_get_module().state_dict())

        # pyre-fixme[16]: `SketchAggregator` has no attribute `cfg`.
        if self.cfg.top_k_ratio > 1.0 or self.cfg.top_k_ratio < 0.0:
            raise AssertionError("top k ratio must be between 0.0 and 1.0.")
        self.num_params = sum(
            [p.numel() for p in global_model.fl_get_module().parameters()]
        )
        self.top_k = int(self.cfg.top_k_ratio * self.num_params)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.reducer, "_target_"):
            cfg.reducer = SketchRoundReducerConfig()

    def step(self) -> Optional[float]:
        """
        Implements FetchSGD steps 11-15.
        """
        reduced_sketch, sum_weights = self.reducer.reduce()

        # momentum
        # pyre-fixme[16]: `SketchAggregator` has no attribute `cfg`.
        self.grad_cs.linear_comb(self.cfg.momentum, reduced_sketch, 1.0)
        # error feedback
        self.error_cs.linear_comb(1.0, self.grad_cs, self.cfg.lr)
        # unsketch
        state_dict = self.error_cs.unsketch_model(k=self.top_k)
        # error accumulation
        topk_cs = clone_count_sketch(self.error_cs)
        topk_cs.sketch_state_dict(state_dict)
        self.error_cs.linear_comb(1.0, topk_cs, -1.0)
        # update global model
        aggregated_module = FLModelParamUtils.clone(self._global_model.fl_get_module())
        FLModelParamUtils.load_state_dict(
            model=aggregated_module, state_dict=state_dict, only_federated_params=False
        )
        FLModelParamUtils.subtract_model(
            minuend=self._global_model.fl_get_module(),
            subtrahend=aggregated_module,
            difference=self._global_model.fl_get_module(),
            only_federated_params=False,
        )

    def init_round(self, reducer: Optional[IFLRoundReducer] = None):
        if reducer is not None and not isinstance(reducer, SketchRoundReducer):
            raise AssertionError(
                f"Must use SketchRoundReducer in SketchAggregator, but got {reducer}"
            )
        if reducer is not None and reducer is not self.reducer:
            self.logger.warning("Changing the round reducer!")
            self.reducer = reducer
        self.reducer.reset(ref_model=self._global_model)


@dataclass
class SketchAggregatorConfig(SyncAggregatorConfig):
    _target_: str = fullclassname(SketchAggregator)
    momentum: float = 0.1
    lr: float = 0.1
    top_k_ratio: float = 1.0
    reducer: SketchRoundReducerConfig = SketchRoundReducerConfig()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.common.pytest_helper import assertEqual, assertRaises
from flsim.compression.em import EM, EmptyClusterResolveError
from flsim.compression.pq import PQ

torch.manual_seed(0)


def test_em_constant() -> None:
    """
    We test our k-means (EM) algorithm on a dummy dataset
    with 3 groups of 100 identical vectors and check that
    the learnt centroids represent the 3 groups.
    """

    # define W as 100 * 3 vectors of dimension 2, the first 100 are
    # filled with 1, the next 100 with 2 and the final 100 with 3
    in_features = 2
    out_features = 100
    seed = torch.ones(out_features, in_features)
    W = torch.cat([seed, 2 * seed, 3 * seed]).t()

    # we quantize with 2 centroids
    em = EM(W, n_centroids=3)
    em.learn()

    # we know the centroids, up to a permutation
    seed = torch.ones(in_features)
    true_centroids = torch.stack([seed, 2 * seed, 3 * seed])
    norm = (em.centroids.sort(dim=0).values - true_centroids).norm()
    assertEqual(norm, 0)


def test_em_error() -> None:
    """
    We try to cluster identical vectors with 3 centroids and check that
    our program raises an error.
    """

    W = torch.ones(2, 500)
    em = EM(W, n_centroids=2)
    with assertRaises(EmptyClusterResolveError):
        em.learn()


def test_pq_linear_constant() -> None:
    """
    We check that PQ on dummy vectors of dimension 2 with a block size 2
    in the linera case yields the expected centroids and that successively
    quantizing and dequantizing amounts to the identity operation.
    """

    # define W as 100 * 3 vectors of dimension 2, the first 100 are
    # filled with 1, the next 100 with 2 and the final 100 with 3
    in_features = 2
    out_features = 100
    seed = torch.ones(out_features, in_features)
    W = torch.cat([seed, 2 * seed, 3 * seed])

    # we quantize with block size 2 (each colun of W is split into 2)
    pq = PQ(W.size(), max_block_size=2, num_codebooks=1, max_num_centroids=3)
    centroids, assignments = pq.encode(W)

    # we now the centroids, up to a permutation
    seed = torch.ones(in_features)
    true_centroids = torch.stack([seed, 2 * seed, 3 * seed])
    norm = (centroids.sort(dim=0).values - true_centroids).norm()

    assertEqual(norm, 0)

    # we should recover exactly W
    diff = (pq.decode(centroids, assignments) - W).norm()
    assertEqual(diff, 0)


def test_pq_conv_constant() -> None:
    """
    We check that PQ on dummy vectors of dimension 9 with a block size 9
    in the convolutional case yields the expected centroids and that
    successively quantizing and dequantizing amounts to the identity operation.
    """

    # define W as 10 * 5 * 3 vectors of dimension 9, the first 100 are
    # filled with 1, the next 100 with 2 and the final 100 with 3
    in_features = 10
    out_features = 5
    kernel_size = 3
    seed = torch.ones(out_features, in_features, kernel_size, kernel_size)
    W = torch.cat([seed, 2 * seed, 3 * torch.ones(4, 10, 3, 3)])

    # we quantize with block size 2 (each colun of W is split into 2)
    pq = PQ(W.size(), max_block_size=9, num_codebooks=1, max_num_centroids=3)
    centroids, assignments = pq.encode(W)

    # we now the centroids, up to a permutation
    seed = torch.ones(kernel_size * kernel_size)
    true_centroids = torch.stack([seed, 2 * seed, 3 * seed])
    norm = (centroids.sort(dim=0).values - true_centroids).norm()
    assertEqual(norm, 0)

    # we should recover exactly W
    diff = (pq.decode(centroids, assignments) - W).norm()
    assertEqual(diff, 0)


def test_pq_many_dicts() -> None:
    """
    We check that PQ on dummy vectors of dimension 4 with a block size 2
    and two learnt codebooks in the linear case yields the expected centroids
    (6 in total, 3 per codebook) and that successively quantizing and
    dequantizing amounts to the identity operation.
    """

    # define W as 300 vectors of dimension 4, the first 100 are filled with
    # 1, the next 100 with 2 and the final 100 with 3
    in_features = 4
    out_features = 100
    seed = torch.ones(out_features, in_features)
    W = torch.cat([seed, 2 * seed, 3 * seed])

    # we quantize with block size 2 (each colun of W is split into 2)
    num_codebooks = 2
    pq = PQ(
        W.size(), max_block_size=2, num_codebooks=num_codebooks, max_num_centroids=3
    )
    centroids, assignments = pq.encode(W)

    # we now the centroids, up to a permutation
    seed = torch.ones(in_features // num_codebooks)
    true_centroids = torch.stack([seed, seed, 2 * seed, 2 * seed, 3 * seed, 3 * seed])
    norm = (centroids.sort(dim=0).values - true_centroids).norm()

    assertEqual(norm, 0)

    # we should recover exactly W
    diff = (pq.decode(centroids, assignments) - W).norm()
    assertEqual(diff, 0)


def test_pq_conv_block_size() -> None:
    """
    We test that the effective block size is 9 when providing a maximum
    block size of 10 in the convolutional case with 1 input feature.
    """

    # define W as 10 * 5 * 3 vectors of dimension 9, the first 100 are
    # filled with 1, the next 100 with 2 and the final 100 with 3
    in_features = 1
    out_features = 5
    kernel_size = 3
    seed = torch.ones(out_features, in_features, kernel_size, kernel_size)
    W = torch.cat([seed, 2 * seed, 3 * seed])

    # we quantize with block size 2 (each colun of W is split into 2)
    pq = PQ(W.size(), max_block_size=10, num_codebooks=1, max_num_centroids=3)
    centroids, assignemnts = pq.encode(W)

    # centroids should have dimension 9, which is the largest acceptable block size
    assertEqual(centroids.size(1), kernel_size * kernel_size)


def test_pq_linear_n_centroids() -> None:
    """
    We check that PQ on 100 random vectors yields a number of centroids equal
    to 16 when providing a larger max_num_centroids equal to 32. Indeed, we
    require n_centroids to be less than n_subvectors // 4 and a power of two.
    """

    # define W as 100 * 3 vectors of dimension 2, the first 100 are
    # filled with 1, the next 100 with 2 and the final 100 with 3
    in_features = 2
    out_features = 100
    W = torch.rand(out_features, in_features)

    # we quantize with block size 2 (each colun of W is split into 2)
    pq = PQ(W.size(), max_block_size=2, num_codebooks=1, max_num_centroids=32)
    centroids, assignments = pq.encode(W)

    # we now the centroids, up to a permutation
    n_centroids = centroids.size(0)
    assertEqual(n_centroids, 16)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import random
from typing import List, Optional

import numpy as np
from flsim.utils.timing.training_duration_distribution import (
    IDurationDistribution,
)


class TrainingTimeEstimator:
    def __init__(
        self,
        total_users: int,
        users_per_round: int,
        epochs: int,
        training_dist: IDurationDistribution,
        num_examples: Optional[List[int]] = None,
    ):
        self.total_users = total_users
        self.users_per_round = users_per_round
        self.epochs = epochs
        self.rounds = int(self.epochs * self.total_users / self.users_per_round)

        self.num_examples: Optional[List[int]] = num_examples
        self.training_dist = training_dist

    def random_select(self):
        """
        Simulate user random selection to return the number of examples for that user
        """
        return 1 if self.num_examples is None else random.choice(self.num_examples)


class SyncTrainingTimeEstimator(TrainingTimeEstimator):
    def __init__(
        self,
        total_users: int,
        users_per_round: int,
        epochs: int,
        training_dist: IDurationDistribution,
        num_examples: Optional[List[int]] = None,
    ):
        super().__init__(
            total_users=total_users,
            users_per_round=users_per_round,
            epochs=epochs,
            num_examples=num_examples,
            training_dist=training_dist,
        )

    def training_time(self):
        """
        Returns training time for SyncFL
        """
        round_completion_time = [
            self.round_completion_time(
                users_per_round=self.users_per_round,
                num_examples=self.num_examples,
                training_dist=self.training_dist,
            )
            for x in range(self.rounds)
        ]
        return sum(round_completion_time)

    def round_completion_time(
        self,
        users_per_round: int,
        num_examples: List[int],
        training_dist: IDurationDistribution,
    ):
        """
        Return the max completion time: straggler effect
        """
        training_times = [
            training_dist.training_duration(self.random_select())
            for _ in range(users_per_round)
        ]
        return max(training_times)


class AsyncTrainingTimeEstimator(TrainingTimeEstimator):
    def __init__(
        self,
        total_users: int,
        users_per_round: int,
        epochs: int,
        training_dist: IDurationDistribution,
        num_examples: Optional[List[int]] = None,
        start_times: Optional[List[int]] = None,
    ):
        super().__init__(
            total_users=total_users,
            users_per_round=users_per_round,
            epochs=epochs,
            num_examples=num_examples,
            training_dist=training_dist,
        )
        self.start_times = start_times

    def training_time(self):
        """
        Returns the training time for AsyncFL
        Assuming client starts training at a linear rate

        """
        user_training_events = self.total_users * self.epochs
        training_durations = [
            self.training_dist.training_duration(self.random_select())
            for _ in range(user_training_events)
        ]
        training_start_times = self.training_start_times(
            training_durations, user_training_events
        )
        training_end_times = self.list_sum(training_start_times, training_durations)
        return max(training_end_times)

    def training_start_times(self, training_durations, user_training_events):
        if self.start_times is None:
            training_start_delta = np.mean(training_durations) / self.users_per_round
            return [
                user_index * training_start_delta
                for user_index in range(user_training_events)
            ]
        else:
            return self.start_times

    @classmethod
    def list_sum(cls, listA: List[float], listB: List[float]):
        ret_val = map(lambda x, y: x + y, listA, listB)
        return list(ret_val)


def get_training_time(
    num_users: int,
    users_per_round: int,
    epochs: int,
    training_dist: IDurationDistribution,
    num_examples: Optional[List[int]] = None,
):
    """
    Returns the estimated training time between SyncFL and AsyncFL
    """
    sync_estimator = SyncTrainingTimeEstimator(
        total_users=num_users,
        users_per_round=users_per_round,
        epochs=epochs,
        num_examples=num_examples,
        training_dist=training_dist,
    )

    async_estimator = AsyncTrainingTimeEstimator(
        total_users=num_users,
        users_per_round=users_per_round,
        epochs=epochs,
        num_examples=num_examples,
        training_dist=training_dist,
    )
    async_time = async_estimator.training_time()
    sync_time = sync_estimator.training_time()
    print(f"Sync {sync_time} Async {async_time}")
    return sync_time, async_time



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import pytest
from flsim.common.pytest_helper import assertEqual, assertAlmostEqual
from flsim.optimizers.sync_aggregators import (
    FedLARSSyncAggregator,
    FedAvgWithLRSyncAggregator,
    FedLARSSyncAggregatorConfig,
    FedAvgWithLRSyncAggregatorConfig,
)
from flsim.tests.utils import MockQuadratic1DFL, Quadratic1D
from omegaconf import OmegaConf


@pytest.fixture(scope="class")
def prepare_lars_optimizer_test(request):
    request.cls.quadratic1D_lars = MockQuadratic1DFL(Quadratic1D())
    request.cls.quadratic1D_sgd = MockQuadratic1DFL(Quadratic1D())


@pytest.mark.usefixtures("prepare_lars_optimizer_test")
class TestLarsOptimizer:
    def test_lars_multiple_steps(self):
        """
            a toy optimization example:
                min f(x) = 100 x^2 - 1

        minima is x=0.0, x is initialized at 1.0.
        """
        # set up quadratic parabola objective and optimizer
        dict_config_lars = {"lr": 0.01, "weight_decay": 0}
        lars_aggregator = FedLARSSyncAggregator(
            **OmegaConf.structured(FedLARSSyncAggregatorConfig(**dict_config_lars)),
            global_model=self.quadratic1D_lars,
        )

        for i in range(10):
            lars_aggregator.optimizer.zero_grad()
            metrics = self.quadratic1D_lars.fl_forward()
            loss_lars = metrics.loss
            loss_lars.backward()

            original_param_value = (
                self.quadratic1D_lars.fl_get_module().state_dict()["x"].item()
            )
            weight_norm = abs(original_param_value)
            lars_aggregator.optimizer.step()
            updated_param_value_lars = (
                self.quadratic1D_lars.fl_get_module().state_dict()["x"].item()
            )

            lars_gradient = list(self.quadratic1D_lars.fl_get_module().parameters())[
                0
            ].grad.data.item()
            lars_gradient_norm = abs(lars_gradient)

            if i == 0:
                assertAlmostEqual(
                    abs((original_param_value - updated_param_value_lars)),
                    0.01,
                    delta=1e-4,
                )
                assertAlmostEqual(lars_gradient_norm, 200, delta=1e-4)

            equivalent_sgd_lr = (
                dict_config_lars["lr"] * weight_norm / lars_gradient_norm
            )
            dict_config_sgd = {"lr": equivalent_sgd_lr}
            sgd_aggregator = FedAvgWithLRSyncAggregator(
                **OmegaConf.structured(
                    FedAvgWithLRSyncAggregatorConfig(**dict_config_sgd)
                ),
                global_model=self.quadratic1D_sgd,
            )

            sgd_aggregator.optimizer.zero_grad()
            metrics = self.quadratic1D_sgd.fl_forward()
            loss_sgd = metrics.loss
            loss_sgd.backward()
            original_param_value = (
                self.quadratic1D_sgd.fl_get_module().state_dict()["x"].item()
            )
            sgd_aggregator.optimizer.step()
            updated_param_value_sgd = (
                self.quadratic1D_sgd.fl_get_module().state_dict()["x"].item()
            )

            assertEqual(updated_param_value_lars, updated_param_value_sgd)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from enum import Enum
from typing import Union

import numpy as np
import torch
from flsim.common.pytest_helper import assertEqual, assertEmpty
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.model import IFLModel
from flsim.optimizers.async_aggregators import (
    AsyncAggregatorConfig,
    FedAdamAsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    FedAdamHybridAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
    create_optimizer_for_async_aggregator,
)
from flsim.optimizers.server_optimizers import (
    FedAdamOptimizerConfig,
    FedAvgWithLROptimizerConfig,
    OptimizerType,
)
from flsim.servers.sync_servers import (
    SyncServerConfig,
)
from flsim.tests.utils import (
    MetricsReporterWithMockedChannels,
    verify_models_equivalent_after_training,
)
from flsim.trainers.async_trainer import AsyncTrainerConfig
from flsim.trainers.sync_trainer import SyncTrainerConfig
from flsim.utils.config_utils import is_target
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.async_trainer_test_utils import (
    create_async_trainer,
    create_event_generator_config,
    get_safe_global_lr,
    run_fl_training,
)
from flsim.utils.tests.helpers.sync_trainer_test_utils import create_sync_trainer
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from flsim.utils.tests.helpers.test_utils import FLTestUtils


class TrainerType(Enum):
    SYNC: str = SyncTrainerConfig._target_.split(".")[-1]
    ASYNC: str = AsyncTrainerConfig._target_.split(".")[-1]
    NONFL: str = "NonFL"


class HybridFLTestUtils:
    @staticmethod
    def get_data_provider(
        num_examples: int,
        num_fl_users: int,
        examples_per_user: int,
        batch_size: int,
        model: IFLModel,
    ) -> IFLDataProvider:
        dummy_dataset = DummyAlphabetDataset(num_examples)
        (
            data_provider,
            fl_data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, examples_per_user, batch_size, model
        )
        assert fl_data_loader.num_total_users == num_fl_users
        return data_provider

    @staticmethod
    def train_comparable_model(
        trainer_to_compare_hybrid_fl_with,
        data_provider,
        global_model,
        server_config,
        local_lr,
        epochs,
        training_rate,
        training_duration_mean,
        training_duration_sd,
    ) -> IFLModel:
        metric_reporter = MetricsReporterWithMockedChannels()
        if trainer_to_compare_hybrid_fl_with == TrainerType.SYNC:
            trainer = create_sync_trainer(
                model=global_model,
                local_lr=local_lr,
                epochs=epochs,
                users_per_round=training_rate,
                server_config=server_config,
            )
            model_to_compare, _ = trainer.train(
                data_provider=data_provider,
                metric_reporter=metric_reporter,
                num_total_users=data_provider.num_users(),
                distributed_world_size=1,
            )
        elif trainer_to_compare_hybrid_fl_with == TrainerType.ASYNC:
            trainer = create_async_trainer(
                model=global_model,
                local_lr=local_lr,
                epochs=epochs,
                aggregator_config=server_config,
                event_generator_config=create_event_generator_config(
                    training_rate=training_rate,
                    training_duration_mean=training_duration_mean,
                    training_duration_sd=training_duration_sd,
                ),
            )
            model_to_compare, _ = trainer.train(
                data_provider=data_provider,
                metric_reporter=metric_reporter,
                num_total_users=data_provider.num_users(),
                distributed_world_size=1,
            )
        elif trainer_to_compare_hybrid_fl_with == TrainerType.NONFL:
            # create an optimizer from aggregator_config.
            # for tests in this file, optimizer will either be
            # torch.optim.SGD or torch.optim.Adam
            if isinstance(server_config, AsyncAggregatorConfig):
                optimizer = create_optimizer_for_async_aggregator(
                    config=server_config,
                    model=global_model.fl_get_module(),
                )
            elif isinstance(server_config, SyncServerConfig):
                optimizer = OptimizerType.create_optimizer(
                    model=global_model.fl_get_module(),
                    config=server_config.server_optimizer,
                )
            else:
                raise AssertionError(f"Incompatible server config:{server_config}")
            model_to_compare, _ = FLTestUtils.train_non_fl(
                data_provider=data_provider,
                global_model=global_model,
                optimizer=optimizer,
                metrics_reporter=metric_reporter,
                epochs=epochs,
            )
        # pyre-fixme[61]: `model_to_compare` may not be initialized here.
        return model_to_compare

    @staticmethod
    def get_hybrid_aggregator(
        aggregator_config, buffer_size, hybrid_lr
    ) -> Union[FedAdamHybridAggregatorConfig, FedAvgWithLRHybridAggregatorConfig]:
        if isinstance(aggregator_config, AsyncAggregatorConfig):
            if "FedAdam" in aggregator_config._target_:
                hybrid_aggregator = FedAdamHybridAggregatorConfig(
                    lr=hybrid_lr,
                    # pyre-ignore[16]
                    weight_decay=aggregator_config.weight_decay,
                    # pyre-ignore[16]
                    eps=aggregator_config.eps,
                    buffer_size=buffer_size,
                )
            else:  # "FedAvgWithLR" in aggregator_config._target_:
                hybrid_aggregator = FedAvgWithLRHybridAggregatorConfig(
                    lr=hybrid_lr,
                    # pyre-ignore[16]
                    momentum=aggregator_config.momentum,
                    buffer_size=buffer_size,
                )
            return hybrid_aggregator
        elif isinstance(aggregator_config, SyncServerConfig):
            if is_target(aggregator_config.server_optimizer, FedAdamOptimizerConfig):
                hybrid_aggregator = FedAdamHybridAggregatorConfig(
                    lr=hybrid_lr,
                    # pyre-ignore[16]
                    weight_decay=aggregator_config.server_optimizer.weight_decay,
                    # pyre-ignore[16]
                    eps=aggregator_config.server_optimizer.eps,
                    buffer_size=buffer_size,
                )
            else:
                hybrid_aggregator = FedAvgWithLRHybridAggregatorConfig(
                    lr=hybrid_lr,
                    # pyre-ignore[16]
                    momentum=aggregator_config.server_optimizer.momentum,
                    buffer_size=buffer_size,
                )
            return hybrid_aggregator
        else:
            raise ValueError("Invalid config", aggregator_config)

    @staticmethod
    def compare_hybrid_fl_same(
        trainer_to_compare_hybrid_fl_with,
        trainer_to_compare_aggregator_config,
        hybrid_aggregator_config,
        base_local_lr,
        hybrid_local_lr,
        epochs,
        num_examples,
        num_fl_users,
        batch_size,
        examples_per_user,
        buffer_size,
        training_rate,
        training_duration_mean,
        training_duration_sd,
    ) -> str:
        # we need to make three copies:
        # to train the model we want to compare with
        global_model = DummyAlphabetFLModel()
        # to train hybrid model
        global_model_hybrid_copy = copy.deepcopy(global_model)
        # to verify training indeed took place
        global_model_init_copy = copy.deepcopy(global_model)

        data_provider = HybridFLTestUtils.get_data_provider(
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            examples_per_user=examples_per_user,
            batch_size=batch_size,
            model=global_model,
        )

        def get_base_trained_model():
            model_to_compare = HybridFLTestUtils.train_comparable_model(
                trainer_to_compare_hybrid_fl_with=trainer_to_compare_hybrid_fl_with,
                data_provider=data_provider,
                global_model=global_model,
                server_config=trainer_to_compare_aggregator_config,
                epochs=epochs,
                local_lr=base_local_lr,
                training_rate=training_rate,
                training_duration_mean=training_duration_mean,
                training_duration_sd=training_duration_sd,
            )
            return model_to_compare

        def get_hybrid_trained_model():
            hybrid_fl_trainer = create_async_trainer(
                model=global_model_hybrid_copy,
                local_lr=hybrid_local_lr,
                epochs=epochs,
                event_generator_config=create_event_generator_config(
                    training_rate=training_rate,
                    training_duration_mean=training_duration_mean,
                    training_duration_sd=training_duration_sd,
                ),
                aggregator_config=hybrid_aggregator_config,
            )

            hybrid_model, _ = hybrid_fl_trainer.train(
                data_provider=data_provider,
                metric_reporter=MetricsReporterWithMockedChannels(),
                num_total_users=data_provider.num_users(),
                distributed_world_size=1,
            )
            return hybrid_model

        hybrid_trained_model = get_hybrid_trained_model()
        base_model = get_base_trained_model()

        error_msg = verify_models_equivalent_after_training(
            base_model,
            hybrid_trained_model,
            global_model_init_copy,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )
        return error_msg

    @staticmethod
    def get_data_params(
        min_num_users, max_num_users, min_examples_per_user, max_examples_per_user
    ):
        r"""
        Generate data parameters for FL training with
        num users in range [min_num_users, max_num_users)
        """
        num_fl_users = np.random.randint(min_num_users, max_num_users)
        examples_per_user = np.random.randint(
            min_examples_per_user, max_examples_per_user
        )
        num_examples = examples_per_user * num_fl_users
        # num_fl_users + 1 because randint upper bound is exclusive
        training_rate = np.random.randint(min_num_users, num_fl_users + 1)
        return num_fl_users, examples_per_user, num_examples, training_rate

    @staticmethod
    def compare_nonfl_hybrid_uneven_data_split(
        total_examples,
        num_fl_users,
        buffer_size,
        non_fl_lr,
        hybrid_global_lr,
        examples_per_user_hybrid,
        examples_per_user_nonfl,
        batch_size_hybrid,
        batch_size_nonfl,
        epochs,
        local_lr=1.0,
    ):
        # to verify training indeed took place
        reference_untrained_model = DummyAlphabetFLModel()
        # to train hybrid model
        hybrid_model = copy.deepcopy(reference_untrained_model)
        # to train nonfl
        nonfl_model = copy.deepcopy(reference_untrained_model)

        nonfl_data_provider = HybridFLTestUtils.get_data_provider(
            num_examples=total_examples,
            num_fl_users=1,
            examples_per_user=examples_per_user_nonfl,
            batch_size=batch_size_nonfl,
            model=nonfl_model,
        )

        hybrid_data_provider = HybridFLTestUtils.get_data_provider(
            num_examples=total_examples,
            num_fl_users=num_fl_users,
            examples_per_user=examples_per_user_hybrid,
            batch_size=batch_size_hybrid,
            model=hybrid_model,
        )

        hybrid_trained_model, _ = run_fl_training(
            fl_model=hybrid_model,
            fl_data_provider=hybrid_data_provider,
            epochs=epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRHybridAggregatorConfig(
                lr=hybrid_global_lr, buffer_size=buffer_size
            ),
            # sequential training training_rate >> training_duration
            training_rate=1,
            training_duration_mean=0,
        )

        optimizer = torch.optim.SGD(
            nonfl_model.fl_get_module().parameters(), lr=non_fl_lr
        )
        nonfl_trained_model, _ = FLTestUtils.train_non_fl(
            data_provider=nonfl_data_provider,
            global_model=nonfl_model,
            optimizer=optimizer,
            epochs=epochs,
        )
        print(f"Local LR {local_lr} Non LR {non_fl_lr} Hybrid LR {hybrid_global_lr}")
        return verify_models_equivalent_after_training(
            nonfl_trained_model,
            hybrid_trained_model,
            reference_untrained_model,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )

    @staticmethod
    def compare_async_hybrid_uneven_data_split(
        total_examples,
        num_fl_users,
        hybrid_num_fl_users,
        async_global_lr,
        hybrid_global_lr,
        batch_size_hybrid,
        batch_size_async,
        epochs,
        training_rate,
        training_duration_mean,
        local_lr=1.0,
    ):
        # to verify training indeed took place
        reference_untrained_model = DummyAlphabetFLModel()
        # to train hybrid model
        hybrid_model = copy.deepcopy(reference_untrained_model)
        # to train async
        async_model = copy.deepcopy(reference_untrained_model)

        async_data_provider = HybridFLTestUtils.get_data_provider(
            num_examples=total_examples,
            num_fl_users=1,
            examples_per_user=total_examples,
            batch_size=batch_size_async,
            model=async_model,
        )
        async_trained_model, _ = run_fl_training(
            fl_model=async_model,
            fl_data_provider=async_data_provider,
            epochs=epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRAsyncAggregatorConfig(lr=async_global_lr),
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
        )

        hybrid_data_provider = HybridFLTestUtils.get_data_provider(
            num_examples=total_examples,
            num_fl_users=num_fl_users,
            examples_per_user=total_examples // num_fl_users,
            batch_size=batch_size_hybrid,
            model=hybrid_model,
        )
        hybrid_trained_model, _ = run_fl_training(
            fl_model=hybrid_model,
            fl_data_provider=hybrid_data_provider,
            epochs=epochs,
            local_lr=local_lr,
            aggregator_config=FedAvgWithLRHybridAggregatorConfig(
                lr=hybrid_global_lr, buffer_size=hybrid_num_fl_users
            ),
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
        )
        print(
            f"Local LR {local_lr} Async LR {async_global_lr} Hybrid LR {hybrid_global_lr}"
        )
        return verify_models_equivalent_after_training(
            async_trained_model,
            hybrid_trained_model,
            reference_untrained_model,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )


class TestFedBuff:
    @classmethod
    def setup_class(cls):
        np.random.seed(0)

    def test_async_hybrid_same_multiple_clients_to_sync(self):
        r"""
        Test when hybrid has multiple clients to sync
        1. Training in parallel. Meaning, both mechanisms start from the same global model (training_time >> training rate)
        2. Total examples is the same for both and local lr is 1.0
        3. Pure Async 1 user with 10 examples should be the same as hybrid 10 users each with 1 example
        """
        # training in parallel training_duration >> training_rate
        training_rate = 1
        training_duration_mean = training_rate * 100

        total_examples = 10
        num_fl_users = 10
        batch_size_hybrid = 1
        batch_size_async = total_examples

        buffer_size = num_fl_users
        local_lr = np.random.sample()
        async_global_lr = np.random.sample()
        hybrid_global_lr = async_global_lr / buffer_size
        error_msg = HybridFLTestUtils.compare_async_hybrid_uneven_data_split(
            total_examples=total_examples,
            num_fl_users=num_fl_users,
            hybrid_num_fl_users=buffer_size,
            async_global_lr=async_global_lr,
            hybrid_global_lr=hybrid_global_lr,
            batch_size_hybrid=batch_size_hybrid,
            batch_size_async=batch_size_async,
            epochs=5,
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
            local_lr=local_lr,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_non_fl_hybrid_same_multiple_clients_to_sync(self):
        """
        Test nonFL and hybrid same with multiple clients to sync

        Both mechanisms train on the number of examples yield the same result if
        1. hybrid is trained sequentially (training rate >> training duration)
        2. NonFL has 1 user with 10 examples and hybrid has 10 users each with 1 example trained for 1 epoch
        3. Local_lr = some random value, global_lr = some random value
        4. Hybrid_global_lr = global_lr / (buffer_size * local_lr)
        """
        total_examples = 10
        num_fl_users = 10
        examples_per_user_hybrid = 1
        batch_size_hybrid = 1
        batch_size_nonfl = total_examples

        buffer_size = 10
        nonfl_lr = np.random.sample()
        hybrid_local_lr = np.random.sample()
        hybrid_global_lr = nonfl_lr / (buffer_size * hybrid_local_lr)

        error_msg = HybridFLTestUtils.compare_nonfl_hybrid_uneven_data_split(
            total_examples=total_examples,
            num_fl_users=num_fl_users,
            examples_per_user_hybrid=examples_per_user_hybrid,
            examples_per_user_nonfl=total_examples,
            batch_size_hybrid=batch_size_hybrid,
            batch_size_nonfl=batch_size_nonfl,
            buffer_size=buffer_size,
            non_fl_lr=nonfl_lr,
            hybrid_global_lr=hybrid_global_lr,
            local_lr=hybrid_local_lr,
            epochs=1,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_async_hybrid_same_sync_every_client(self):
        r"""
        Hybrid and Async should yield the same model:
        1. Async and Hybrid training duration distribution are the same
        training_duration_sd should be small
        2. Hybrid update global model every 1 client
        """
        # random learning rate between 0 and 10
        local_lr = np.random.sample()
        global_lr = np.random.sample() * 10
        buffer_size = 1
        for base_aggregator_config in [
            FedAvgWithLRAsyncAggregatorConfig(lr=global_lr),
            FedAdamAsyncAggregatorConfig(lr=global_lr, eps=1e-2),
        ]:
            for batch_size in [4, 16, 32]:
                (
                    num_fl_users,
                    examples_per_user,
                    num_examples,
                    training_rate,
                ) = HybridFLTestUtils.get_data_params(
                    min_num_users=2,
                    max_num_users=10,
                    min_examples_per_user=1,
                    max_examples_per_user=10,
                )
                error_msg = HybridFLTestUtils.compare_hybrid_fl_same(
                    trainer_to_compare_hybrid_fl_with=TrainerType.ASYNC,
                    trainer_to_compare_aggregator_config=base_aggregator_config,
                    hybrid_aggregator_config=HybridFLTestUtils.get_hybrid_aggregator(
                        aggregator_config=base_aggregator_config,
                        buffer_size=buffer_size,
                        hybrid_lr=global_lr,
                    ),
                    hybrid_local_lr=local_lr,
                    base_local_lr=local_lr,
                    epochs=5,
                    num_examples=num_examples,
                    num_fl_users=num_fl_users,
                    batch_size=batch_size,
                    examples_per_user=examples_per_user,
                    buffer_size=buffer_size,
                    training_rate=training_rate,
                    training_duration_mean=1,
                    training_duration_sd=0,
                )
                assertEqual(error_msg, "")

    def test_nonfl_hybrid_same_sgd(self):
        r"""
        Hybrid and NonFL should yield the same model:
        1. Hybrid's training rate = 1, training duration ~ N(0, 0)
        2. Round robin user selector
        3. Hybrid takes global step  every 1 clients
        """
        (
            num_fl_users,
            examples_per_user,
            num_examples,
            _,
        ) = HybridFLTestUtils.get_data_params(
            min_num_users=1,
            max_num_users=10,
            min_examples_per_user=1,
            max_examples_per_user=10,
        )
        local_lr = 1.0
        buffer_size = 1
        for batch_size in [4, 16, 32]:
            global_lr = get_safe_global_lr(batch_size, examples_per_user)
            base_aggregator_config = SyncServerConfig(
                server_optimizer=FedAvgWithLROptimizerConfig(lr=global_lr, momentum=0.0)
            )
            error = HybridFLTestUtils.compare_hybrid_fl_same(
                trainer_to_compare_hybrid_fl_with=TrainerType.NONFL,
                trainer_to_compare_aggregator_config=base_aggregator_config,
                hybrid_aggregator_config=HybridFLTestUtils.get_hybrid_aggregator(
                    aggregator_config=base_aggregator_config,
                    buffer_size=buffer_size,
                    hybrid_lr=global_lr,
                ),
                hybrid_local_lr=local_lr,
                base_local_lr=local_lr,
                epochs=1,
                num_examples=num_examples,
                num_fl_users=num_fl_users,
                batch_size=batch_size,
                examples_per_user=examples_per_user,
                buffer_size=1,
                training_rate=1,
                training_duration_mean=0,
                training_duration_sd=0,
            )
            assertEqual(error, "")

    def test_nonfl_hybrid_same_adam(self):
        r"""
        Hybrid and NonFL should yield the same model:
        1. Hybrid's training rate = 1, training duration ~ N(0, 0)
        2. Round robin user selector
        3. Hybrid takes global step  every 1 clients
        """
        (
            num_fl_users,
            examples_per_user,
            num_examples,
            _,
        ) = HybridFLTestUtils.get_data_params(
            min_num_users=1,
            max_num_users=10,
            min_examples_per_user=1,
            max_examples_per_user=10,
        )
        local_lr = 1.0
        global_lr = np.random.sample() * 0.01
        base_aggregator_config = SyncServerConfig(
            server_optimizer=FedAdamOptimizerConfig(lr=global_lr, eps=1e-2)
        )
        buffer_size = 1
        error = HybridFLTestUtils.compare_hybrid_fl_same(
            trainer_to_compare_hybrid_fl_with=TrainerType.NONFL,
            trainer_to_compare_aggregator_config=base_aggregator_config,
            hybrid_aggregator_config=HybridFLTestUtils.get_hybrid_aggregator(
                aggregator_config=base_aggregator_config,
                buffer_size=buffer_size,
                hybrid_lr=global_lr,
            ),
            hybrid_local_lr=local_lr,
            base_local_lr=local_lr,
            epochs=5,
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            batch_size=examples_per_user,
            examples_per_user=examples_per_user,
            buffer_size=1,
            training_rate=1,
            training_duration_mean=0,
            training_duration_sd=0,
        )
        assertEqual(error, "")

    def test_sync_hybrid_same_sgd(self):
        r"""
        Hybrid and Sync should yield the same model:
        1. For simplification, assume training_rate = total_users.
        Without this constraint, Hybrid==Sync only for 1 round of sync.
        2. `(Training_rate*num_users) << training_duration_mean`.
        E.g, `#users=10, training rate = 1, training_duration_mean > 10`.
        This is needed so that all users train in parallel.
        In particular, in a single epoch, every user starts with the same initial model
        3. Round robin user selector
        4. Set hybrid_global_lr = sync_global_lr / num_clients_sync
        """
        (
            num_fl_users,
            examples_per_user,
            num_examples,
            _,
        ) = HybridFLTestUtils.get_data_params(
            min_num_users=1,
            max_num_users=5,
            min_examples_per_user=1,
            max_examples_per_user=5,
        )
        local_lr = 1.0
        global_lr = 1.0
        base_aggregator_config = SyncServerConfig(
            server_optimizer=FedAvgWithLROptimizerConfig(lr=global_lr, momentum=0.0)
        )
        buffer_size = num_fl_users
        for batch_size in [4, 16, 32]:
            print(f"{num_fl_users} {examples_per_user} {local_lr} {global_lr}")
            error = HybridFLTestUtils.compare_hybrid_fl_same(
                trainer_to_compare_hybrid_fl_with=TrainerType.SYNC,
                trainer_to_compare_aggregator_config=base_aggregator_config,
                hybrid_aggregator_config=HybridFLTestUtils.get_hybrid_aggregator(
                    aggregator_config=base_aggregator_config,
                    buffer_size=buffer_size,
                    hybrid_lr=global_lr / buffer_size,
                ),
                hybrid_local_lr=local_lr,
                base_local_lr=local_lr,
                epochs=1,
                num_examples=num_examples,
                num_fl_users=num_fl_users,
                batch_size=batch_size,
                examples_per_user=examples_per_user,
                buffer_size=buffer_size,
                training_rate=buffer_size,
                training_duration_mean=buffer_size * 2,
                training_duration_sd=0,
            )
            assertEqual(error, "")

    def test_sync_hybrid_same_adam(self):
        r"""
        For sync == hybrid adam,
        `hybrid_local_lr = sync_local_lr / buffer_size`
        and hybrid and sync global lr's should be the same.

        Sync and hybrid compute different "delta"
        hybrid_delta = sync_delta * buffer_size
        to fix this difference, for SGD, we set `global_lr_hybrid = global_lr_sync/buffer_size`
        however, for Adam, this normalization doesn't work:
        Since Adam stores first and second moments (mean and variance) of deltas.
        In particular, the following are not equivalent:
        delta=d, lr=l
        delta=d*k, lr=l/k
        instead, we have to set `local_lr_hybrid = local_lr_sync / buffer_size`
        """
        (
            num_fl_users,
            examples_per_user,
            num_examples,
            _,
        ) = HybridFLTestUtils.get_data_params(
            min_num_users=1,
            max_num_users=10,
            min_examples_per_user=1,
            max_examples_per_user=10,
        )

        local_lr = 1.0
        global_lr = np.random.sample() * 0.01
        base_aggregator_config = SyncServerConfig(
            server_optimizer=FedAdamOptimizerConfig(lr=global_lr, eps=1e-2)
        )
        buffer_size = num_fl_users
        error = HybridFLTestUtils.compare_hybrid_fl_same(
            trainer_to_compare_hybrid_fl_with=TrainerType.SYNC,
            trainer_to_compare_aggregator_config=base_aggregator_config,
            hybrid_aggregator_config=HybridFLTestUtils.get_hybrid_aggregator(
                aggregator_config=base_aggregator_config,
                buffer_size=buffer_size,
                hybrid_lr=global_lr,
            ),
            hybrid_local_lr=local_lr / num_fl_users,
            base_local_lr=local_lr,
            epochs=1,
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            batch_size=examples_per_user,
            examples_per_user=examples_per_user,
            buffer_size=buffer_size,
            training_rate=buffer_size,
            training_duration_mean=buffer_size * 100,
            training_duration_sd=0,
        )
        print(f"{num_fl_users} {examples_per_user} {local_lr} {global_lr}")
        assertEqual(error, "")

    def test_partial_model_update(self):
        r"""
        Test for partial update

        Assume we have 2 users (user1 and user2), buffer_size = 2, and both nonfl and fl training have
        exactly one batch training sequentially (training_rate >> training_duration), and denote initial global model as g0
        with the following timeline

        user1 starts training (receives g0)
        user1 ends training
        user2 starts training (note: user2 should receive g0)
        user2 ends training

        this sequence should produce the same model as non-fl training on user1 + user2 datasets
        """
        num_fl_users = 2
        total_examples = 20
        examples_per_user_hybrid = total_examples // num_fl_users
        buffer_size = num_fl_users

        nonfl_lr = np.random.sample()
        hybrid_local_lr = np.random.sample()
        # We need to normalize global_lr to account for the local training
        # hence we need to divide by the (buffer_size * hybrid_local_lr)
        hybrid_global_lr = nonfl_lr / (buffer_size * hybrid_local_lr)

        error_msg = HybridFLTestUtils.compare_nonfl_hybrid_uneven_data_split(
            total_examples=total_examples,
            num_fl_users=num_fl_users,
            buffer_size=buffer_size,
            non_fl_lr=nonfl_lr,
            hybrid_global_lr=hybrid_global_lr,
            local_lr=hybrid_local_lr,
            examples_per_user_hybrid=examples_per_user_hybrid,
            examples_per_user_nonfl=total_examples,
            batch_size_hybrid=examples_per_user_hybrid,
            batch_size_nonfl=total_examples,
            epochs=1,
        )
        print(
            f"NonFL LR {nonfl_lr} Hybrid Local LR {hybrid_local_lr}  Hybrid Global LR {hybrid_global_lr}"
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_remaining_clients_to_sync(self):
        """
        An aggregator can have unaggregated clients before
        an end of the epoch

        This case can happen in two scenario
        1. clients_to_sync > total_users
        2. total_users is not divisible by clients_to_sync (num_fl_users % clients_to_sync != 0)
        """
        # training in parallel training_duration >> training_rate
        training_rate = 1
        training_duration_mean = training_rate * 100

        total_examples = 10
        num_fl_users = 10
        batch_size_hybrid = 1
        batch_size_async = total_examples

        # test for scenario 1
        buffer_size = num_fl_users * 10
        local_lr = np.random.sample()
        async_global_lr = np.random.sample()

        hybrid_global_lr = async_global_lr / num_fl_users
        error_msg = HybridFLTestUtils.compare_async_hybrid_uneven_data_split(
            total_examples=total_examples,
            num_fl_users=num_fl_users,
            hybrid_num_fl_users=buffer_size,
            async_global_lr=async_global_lr,
            hybrid_global_lr=hybrid_global_lr,
            batch_size_hybrid=batch_size_hybrid,
            batch_size_async=batch_size_async,
            epochs=1,
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
            local_lr=local_lr,
        )
        assertEmpty(error_msg, msg=error_msg)

        # test for scenario 2
        buffer_size = 4
        hybrid_global_lr = async_global_lr / num_fl_users
        error_msg = HybridFLTestUtils.compare_async_hybrid_uneven_data_split(
            total_examples=total_examples,
            num_fl_users=num_fl_users,
            hybrid_num_fl_users=buffer_size,
            async_global_lr=async_global_lr,
            hybrid_global_lr=hybrid_global_lr,
            batch_size_hybrid=batch_size_hybrid,
            batch_size_async=batch_size_async,
            epochs=1,
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
            local_lr=local_lr,
        )
        assertEmpty(error_msg, msg=error_msg)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import List

from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from hydra.utils import instantiate
from omegaconf import MISSING


class TimeOutSimulator(abc.ABC):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=TimeOutSimulatorConfig,
            **kwargs,
        )

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def simulate_per_example_training_time(self) -> float:
        """
        return avg unit of time required to process an example
        """
        pass

    @abc.abstractmethod
    def simulate_training_time(self, device_perf: float, num_samples: int) -> float:
        """
        return training time on one client
        """
        pass

    @abc.abstractmethod
    def track_training_time_distribution(self, one_user_training_time: float) -> None:
        """
        update the empirical training time statistics
        """
        pass

    @abc.abstractmethod
    def track_fl_elapsed_time(self, training_time_in_round: List[float]) -> None:
        """
        track total unit of time taken by FL so far given list of user training time
        in the round
        """
        pass

    @abc.abstractmethod
    def user_timeout(self, training_time: float) -> bool:
        """
        time out for one user
        """
        pass

    @abc.abstractmethod
    def stop_fl(self) -> bool:
        """
        stopping condition for entire FL training
        """
        pass

    @property
    @abc.abstractmethod
    def sample_mean_per_user(self) -> float:
        """
        empirical mean of training time per user
        """
        pass

    @property
    @abc.abstractmethod
    def sample_var_per_user(self) -> float:
        """
        empirical variance of training time per user
        """
        pass


class NeverTimeOutSimulator(TimeOutSimulator):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=NeverTimeOutSimulatorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)
        self._sample_mean_per_user = 0.0
        self._sample_var_per_user = 0.0

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def simulate_per_example_training_time(self) -> float:
        """
        training finish instantaneously
        """
        return 0.0

    def simulate_training_time(self, device_perf: float, num_samples: int) -> float:
        """
        training finish instantaneously
        """
        return 0.0

    def track_training_time_distribution(self, one_user_training_time: float) -> None:
        pass

    def user_timeout(self, training_time: float) -> bool:
        return False

    @property
    def sample_mean_per_user(self) -> float:
        return 0.0

    @property
    def sample_var_per_user(self) -> float:
        return 0.0

    def track_fl_elapsed_time(self, training_time_in_round: List[float]) -> None:
        pass

    def stop_fl(self) -> bool:
        return False


class GaussianTimeOutSimulator(TimeOutSimulator):
    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=GaussianTimeOutSimulatorConfig,
            **kwargs,
        )

        super().__init__(**kwargs)

        self.duration_distribution_generator = instantiate(
            self.cfg.duration_distribution_generator
        )
        self._num_users_tracked = 0
        self._num_users_succeed = 0
        self._sample_mean_per_user = 0.0
        self._sample_sec_moment_per_user = 0.0
        self._sample_var_per_user = 0.0
        self._fl_stopping_time = self.cfg.fl_stopping_time
        self._fl_total_elapse_time = 0.0

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @property
    def fl_stopping_time(self):
        return self._fl_stopping_time

    @property
    def sample_mean_per_user(self):
        return self._sample_mean_per_user

    @property
    def sample_var_per_user(self):
        return self._sample_var_per_user

    def simulate_per_example_training_time(self) -> float:
        """
        return an gaussian R.V. representing # unit of time per example
        """
        return self.duration_distribution_generator.bounded_gaussian_sample()

    def simulate_training_time(self, device_perf: float, num_samples: int) -> float:
        """
        return training time on one client
        """
        # pyre-fixme[16]: `GaussianTimeOutSimulator` has no attribute `cfg`.
        return min([device_perf * num_samples, self.cfg.timeout_wall_per_round])

    def track_training_time_distribution(self, one_user_training_time: float) -> None:
        self._num_users_tracked += 1
        # update sample mean
        self._sample_mean_per_user = (
            self._sample_mean_per_user * (self._num_users_tracked - 1)
            + one_user_training_time
        ) / self._num_users_tracked
        # update sample second moment
        self._sample_sec_moment_per_user = (
            self._sample_sec_moment_per_user * (self._num_users_tracked - 1)
            + one_user_training_time ** 2
        ) / self._num_users_tracked
        # update sample variance, with degree of freedom correction
        if self._num_users_tracked > 1:
            self._sample_var_per_user = (
                self._num_users_tracked / (self._num_users_tracked - 1)
            ) * (self._sample_sec_moment_per_user - self._sample_mean_per_user ** 2)

    def user_timeout(self, training_time: float) -> bool:
        """
        training time out for one user
        """
        # pyre-fixme[16]: `GaussianTimeOutSimulator` has no attribute `cfg`.
        return training_time >= self.cfg.timeout_wall_per_round

    def stop_fl(self) -> bool:
        """
        stop entire FL training when total budget exceeds
        """
        return self._fl_total_elapse_time >= self._fl_stopping_time

    def track_fl_elapsed_time(self, training_time_in_round: List[float]) -> None:
        self._fl_total_elapse_time += min(
            # pyre-fixme[16]: `GaussianTimeOutSimulator` has no attribute `cfg`.
            [self.cfg.timeout_wall_per_round, max(training_time_in_round)]
        )


@dataclass
class TimeOutSimulatorConfig:
    _target_: str = MISSING
    _recursive_: bool = False


@dataclass
class NeverTimeOutSimulatorConfig(TimeOutSimulatorConfig):
    _target_: str = fullclassname(NeverTimeOutSimulator)


@dataclass
class GaussianTimeOutSimulatorConfig(TimeOutSimulatorConfig):
    _target_: str = fullclassname(GaussianTimeOutSimulator)
    timeout_wall_per_round: float = 1.0
    fl_stopping_time: float = 1.0
    duration_distribution_generator: PerExampleGaussianDurationDistributionConfig = (
        PerExampleGaussianDurationDistributionConfig()
    )
